__author__ = 'luckydonald'

import unittest
from datetime import datetime
from typing import get_type_hints
from typing import Optional, Union, Any, Type, List, Tuple, Dict

from fastorm import FastORM
from tests.tools_for_the_tests_of_fastorm import extract_sql_from_docstring


class Table1(FastORM):
    """
        INSERT INTO "table1" ("name","number")
         VALUES ($1,$2)
         RETURNING "id"
        ;
    """
    _table_name = 'table1'
    _automatic_fields = ['id']
    _primary_keys = ['id']

    id: Union[int, None]
    name: str
    number: int
# end class


# ----------------------------------------------------


class InsertRowTestCase(unittest.TestCase):
    def test_insert(self):
        expected_sql = extract_sql_from_docstring(Table1)
        expected_params = ["Sample Text", 123]
        row = Table1(
            name=expected_params[0],
            number=expected_params[1]
        )
        actual_sql, *actual_params = row.build_sql_insert()
        self.assertEqual(expected_sql, actual_sql)
        self.assertEqual(expected_params, actual_params)
    # end def
# end class


if __name__ == '__main__':
    unittest.main()
# end if




import unittest
from datetime import datetime
from typing import Optional, Union, Any, Type, List, Tuple, Dict
from pydantic import dataclasses, BaseModel
from pydantic.fields import ModelField, Undefined, Field

from fastorm import FastORM, Autoincrement, FieldInfo
from fastorm.compat import get_type_hints_with_annotations
from tests.tools_for_the_tests_of_fastorm import extract_sql_from_docstring, VerboseTestCase


@dataclasses.dataclass
class ExpectedResult(object):
    is_optional: bool
    sql_type: str
    default: Any
# end class

ExpectedResult: Type[Any]


class OtherTable(FastORM):
    """
        CREATE TABLE "other_table" (
          "id_part_1" BIGINT NOT NULL,
          "id_part_2" TEXT NOT NULL,
          PRIMARY KEY ("id_part_1", "id_part_2")
        );
    """
    _table_name = 'other_table'
    _primary_keys = ['id_part_1', 'id_part_2']
    _ignored_fields = []

    id_part_1: int
    id_part_2: str
# end class


class TheReferenceHasBeenDoubled(FastORM):
    """
        CREATE TABLE "double_reference" (
          "another_reference__id_part_1" BIGINT NOT NULL,
          "another_reference__id_part_2" TEXT NOT NULL,
          PRIMARY KEY ("another_reference__id_part_1", "another_reference__id_part_2")
        );
        CREATE INDEX "idx_double_reference___another_reference__id_part_1" ON "double_reference" ("another_reference__id_part_1");
        CREATE INDEX "idx_double_reference___another_reference__id_part_2" ON "double_reference" ("another_reference__id_part_2");
        ALTER TABLE "double_reference" ADD CONSTRAINT "fk_double_reference___another_reference__id_part_1" FOREIGN KEY ("another_reference__id_part_1") REFERENCES "other_table" ("id_part_1") ON DELETE CASCADE;
        ALTER TABLE "double_reference" ADD CONSTRAINT "fk_double_reference___another_reference__id_part_2" FOREIGN KEY ("another_reference__id_part_2") REFERENCES "other_table" ("id_part_2") ON DELETE CASCADE;
    """
    _table_name = 'double_reference'
    _primary_keys = ['another_reference']
    _ignored_fields = []

    another_reference: OtherTable
# end class


class SystemUnderTest(FastORM):
    """
        CREATE TABLE "cool_table_yo" (
          "t0_id" BIGSERIAL NOT NULL PRIMARY KEY,
          "t1_1" TEXT NOT NULL,
          "t1_2" TEXT NOT NULL,
          "t1_3" TEXT NOT NULL,
          "t1_4" TEXT NOT NULL,
          "t2_1" TEXT,
          # "t2_2" TEXT,
          # "t2_3" TEXT,
          "t2_4" TEXT,
          "t2_5" TEXT,
          "t2_6" TEXT,
          # "t2_7" TEXT,
          # "t2_8" TEXT,
          # "t2_9" TEXT,
          "t3_1" TEXT NOT NULL,
          "t3_2" BIGINT NOT NULL,
          "t3_3" DOUBLE PRECISION NOT NULL,
          "t3_4" BOOLEAN NOT NULL,
          "t5_1" TIMESTAMP NOT NULL,
          "t6_1" TEXT NOT NULL DEFAULT 'test',
          "t6_2" TEXT DEFAULT NULL,
          "t6_3" BIGINT NOT NULL DEFAULT 69,
          "t6_4" TEXT NOT NULL DEFAULT 'this test will proof if "something" ain''t escaped properly. ^^''',
          "t7_1__id_part_1" BIGINT NOT NULL,
          "t7_1__id_part_2" TEXT NOT NULL,
          "t8_1" JSONB NOT NULL,
          "t8_2" BIGINT[] NOT NULL,
          "t8_3" BIGINT[][] NOT NULL,
          "t8_4" JSONB NOT NULL,
          "t9_1" BIGINT[] NOT NULL,
          "t9_2" JSONB NOT NULL
        );
        CREATE INDEX "idx_cool_table_yo___t7_1__id_part_1" ON "cool_table_yo" ("t7_1__id_part_1");
        CREATE INDEX "idx_cool_table_yo___t7_1__id_part_2" ON "cool_table_yo" ("t7_1__id_part_2");
        ALTER TABLE "cool_table_yo" ADD CONSTRAINT "fk_cool_table_yo___t7_1__id_part_1" FOREIGN KEY ("t7_1__id_part_1") REFERENCES "other_table" ("id_part_1") ON DELETE CASCADE;
        ALTER TABLE "cool_table_yo" ADD CONSTRAINT "fk_cool_table_yo___t7_1__id_part_2" FOREIGN KEY ("t7_1__id_part_2") REFERENCES "other_table" ("id_part_2") ON DELETE CASCADE;
        """
    _table_name = 'cool_table_yo'
    _primary_keys = ['t0_id']
    _automatic_fields = ['t0_id']
    _ignored_fields = []

    #
    # Required str
    #

    t0_id: int = Field(default_factory=Autoincrement)
    __result__t0_id = ExpectedResult(is_optional=False, sql_type="BIGINT", default=Undefined)

    t1_1: str
    __result__t1_1 = ExpectedResult(is_optional=False, sql_type="TEXT", default=Undefined)

    t1_2: Union[str]
    __result__t1_2 = ExpectedResult(is_optional=False, sql_type="TEXT", default=Undefined)

    t1_3: Union[Union[str]]
    __result__t1_3 = ExpectedResult(is_optional=False, sql_type="TEXT", default=Undefined)

    t1_4: Union[str, str]
    __result__t1_4 = ExpectedResult(is_optional=False, sql_type="TEXT", default=Undefined)


    #
    # Optional str
    #

    t2_1: Union[str, None]
    __result__t2_1 = ExpectedResult(is_optional=True, sql_type="TEXT", default=Undefined)

    # t2_2: str | None
    __result__t2_2 = ExpectedResult(is_optional=True, sql_type="TEXT", default=Undefined)

    # t2_3: None | str
    __result__t2_3 = ExpectedResult(is_optional=True, sql_type="TEXT", default=Undefined)

    t2_4: Optional[str]
    __result__t2_4 = ExpectedResult(is_optional=True, sql_type="TEXT", default=Undefined)

    t2_5: Union[Union[str, None], None]
    __result__t2_5 = ExpectedResult(is_optional=True, sql_type="TEXT", default=Undefined)

    t2_6: Union[Optional[str], None]
    __result__t2_6 = ExpectedResult(is_optional=True, sql_type="TEXT", default=Undefined)

    # t2_7: Union[str | None, None]
    __result__t2_7 = ExpectedResult(is_optional=True, sql_type="TEXT", default=Undefined)

    # t2_8: Union[str | None] | None
    __result__t2_8 = ExpectedResult(is_optional=True, sql_type="TEXT", default=Undefined)

    # t2_9: Union[str | None] | Optional[None] | None
    __result__t2_9 = ExpectedResult(is_optional=True, sql_type="TEXT", default=Undefined)

    #
    # base types
    #

    t3_1: str
    __result__t3_1 = ExpectedResult(is_optional=False, sql_type="TEXT", default=Undefined)

    t3_2: int
    __result__t3_2 = ExpectedResult(is_optional=False, sql_type="BIGINT", default=Undefined)

    t3_3: float
    __result__t3_3 = ExpectedResult(is_optional=False, sql_type="DOUBLE PRECISION", default=Undefined)

    t3_4: bool
    __result__t3_4 = ExpectedResult(is_optional=False, sql_type="BOOLEAN", default=Undefined)

    #
    # more uncommon types
    #

    t5_1: datetime
    __result__t5_1 = ExpectedResult(is_optional=False, sql_type="TIMESTAMP", default=Undefined)

    #
    # defaults
    #

    t6_1: str = "test"
    __result__t6_1 = ExpectedResult(is_optional=False, sql_type="TEXT", default="test")

    t6_2: Optional[str] = None
    __result__t6_2 = ExpectedResult(is_optional=True, sql_type="TEXT", default=None)

    t6_3: int = 69
    __result__t6_3 = ExpectedResult(is_optional=False, sql_type="BIGINT", default=69)

    t6_4: str = "this test will proof if \"something\" ain't escaped properly. ^^'"
    __result__t6_4 = ExpectedResult(is_optional=False, sql_type="TEXT", default="this test will proof if \"something\" ain't escaped properly. ^^'")

    #
    # references
    #

    t7_1: OtherTable
    # For the single field FastORM.match_type(…) this results as JSONB (because BaseModel)
    # For the table view FastORM.build_sql_create() this results as two keys with the respective types.
    __result__t7_1 = ExpectedResult(is_optional=False, sql_type="JSONB", default=Undefined)

    #
    # special cases for lists
    #

    t8_1: list
    __result__t8_1 = ExpectedResult(is_optional=False, sql_type="JSONB", default=Undefined)

    t8_2: List[int]
    __result__t8_2 = ExpectedResult(is_optional=False, sql_type="BIGINT[]", default=Undefined)

    t8_3: List[List[int]]
    __result__t8_3 = ExpectedResult(is_optional=False, sql_type="BIGINT[][]", default=Undefined)

    t8_4: List[List[Union[str, int]]]
    __result__t8_4 = ExpectedResult(is_optional=False, sql_type="JSONB", default=Undefined)

    #
    # special cases for tuples
    #

    t9_1: Tuple[int, int, int]
    __result__t9_1 = ExpectedResult(is_optional=False, sql_type="BIGINT[]", default=Undefined)

    t9_2: Tuple[int, str, float]
    __result__t9_2 = ExpectedResult(is_optional=False, sql_type="JSONB", default=Undefined)

# end class


class WrongStuff(BaseModel):
    #
    # wrong stuff
    #
    t4_1: Union[int, str]
    __result__t4_1 = TypeError

    t4_2: Optional[Union[int, str]]
    __result__t4_2 = TypeError

    t4_3: Union[int, str]
    __result__t4_3 = TypeError

    t4_4: Union[int, str, None]
    __result__t4_4 = TypeError

    # t4_5: str = None
    __result__t4_5 = ValueError
# end class


class CreateTableTestCase(VerboseTestCase):
    def test_type_detection_typing(self):
        type_hints: Dict[str, any] = get_type_hints_with_annotations(SystemUnderTest)
        for key, type_hint in type_hints.items():
            if key.startswith('_'):
                continue
            # end if
            expected_result: ExpectedResult | Type[Exception]
            expected_result = getattr(SystemUnderTest, f'_{SystemUnderTest.__name__}__result__{key}')
            with self.subTest(msg=key):
                print(key, ",", type_hint, ",", expected_result)
                if isinstance(expected_result, ExpectedResult):
                    is_optional, sql_type = FastORM.match_type(type_hint=type_hint, key=key)
                    self.assertEqual(expected_result.sql_type, sql_type)
                    self.assertEqual(expected_result.is_optional, is_optional)
                else:
                    with self.assertRaises(expected_result):
                        is_optional, sql_type = FastORM.match_type(type_hint=type_hint)
                        print("failed assertion!", key, is_optional, sql_type)
                    # end with
            # end which
        # end for
    # end def

    def test_type_detection_pydantic(self):
        type_hints: Dict[str, FieldInfo[ModelField]] = SystemUnderTest.get_fields_typehints()
        for key, type_hint in type_hints.items():
            expected_result: ExpectedResult | Type[Exception]
            expected_result = getattr(SystemUnderTest, f'_{SystemUnderTest.__name__}__result__{key}')
            with self.subTest(msg=key):
                print(key, ",", type_hint, ",", expected_result)
                if isinstance(expected_result, ExpectedResult):
                    is_optional, sql_type = FastORM.match_type(type_hint=type_hint.types[0].type_, key=key)
                    self.assertEqual(expected_result.sql_type, sql_type, msg='sql_type')
                    self.assertEqual(expected_result.is_optional, is_optional, msg='is_optional')
                else:
                    with self.assertRaises(expected_result):
                        is_optional, sql_type = FastORM.match_type(type_hint=type_hint)
                        print("failed assertion!", key, is_optional, sql_type)
                    # end with
            # end which
        # end for
    # end def

    def test_sql_text(self):
        self.maxDiff = None
        for table_cls in (OtherTable, TheReferenceHasBeenDoubled, SystemUnderTest):
            with self.subTest(msg=f'class {table_cls.__name__}'):
                expected_sql = extract_sql_from_docstring(table_cls)
                actual_sql, *actual_params = table_cls.build_sql_create()
                self.assertEqual(expected_sql, actual_sql)
                self.assertListEqual([], actual_params)
            # end with
        # end for
    # end def

    def test_TYPES_mapping_subclass_shadowing(self):
        """
        if you have `class A(object): pass` and `class B(A): pass`
        if you would list `A` first, it will never reach `B` in processing as `A` already matches all `B` objects..
        :return:
        """

        LISTS_TO_CHECK = {
            "COLUMN_TYPES": FastORM._COLUMN_TYPES,
            "COLUMS_AUTO_TYPES": FastORM._COLUMN_AUTO_TYPES
        }
        for list_name, list_to_check in LISTS_TO_CHECK.items():
            classes = list(list_to_check.keys())
            with self.subTest(msg=f'{FastORM.__name__}.{list_name}'):
                for first_class_index, first_class in enumerate(classes):
                    # for all classes which are after first_class(_index)
                    for second_class_index, second_class in enumerate(classes[first_class_index + 1:], start=first_class_index + 1):
                        # the second may not be a subclass of the first, that would mean the fist would shadow the second.
                        self.assertFalse(
                            issubclass(second_class, first_class),
                            msg=f'The broader class {first_class!r} (at index {first_class_index}) is shadowing the later class {second_class!r} (at index {second_class_index}).')
                    # end for
                # end for
            # end with
        # end for
    # end def
# end class


if __name__ == '__main__':
    unittest.main()
# end if



import unittest
from datetime import datetime
from textwrap import dedent
from typing import get_type_hints
from typing import Optional, Union, Any, Type, List, Tuple, Dict
from pydantic import dataclasses, BaseModel
from pydantic.fields import ModelField, Undefined, Field

from fastorm import FastORM
from fastorm.compat import get_type_hints_with_annotations
from tests.tools_for_the_tests_of_fastorm import extract_sql_from_docstring


class OtherTable(FastORM):
    """
        CREATE TABLE "other_table" (
          "id_part_1" TEXT NOT NULL,
          "id_part_2" BIGINT NOT NULL,
          PRIMARY KEY ("id_part_1", "id_part_2")
        );
    """
    _table_name = 'cool_table_name_yoooo'
    _primary_keys = ['id_part_1', 'id_part_2']
    _ignored_fields = []

    id_part_1: int
    id_part_2: str
    foo: float
    bar: str
# end class


class UpdateRowTestCase(unittest.TestCase):
    def test_foo(self):
        object = OtherTable(
            id_part_1=12,
            id_part_2='banana',
            foo=2.35,
            bar="kiwi",
        )
        self.assertDictEqual(
            {},
            object._database_cache,
            "database cache starts empty"
        )

        self.assertDictEqual(
            {'id_part_1': 12, 'id_part_2': 'banana', 'foo': 2.35, 'bar': 'kiwi'},
            object.get_changes(),
            "get_changes() identifies initial changes correctly"
        )

        object._database_cache_overwrite_with_current()

        self.assertDictEqual(
            {},
            object.get_changes(),
            "get_changes() identifies no changes after they were synced"
        )

        self.assertDictEqual(
            {'id_part_1': 12, 'id_part_2': 'banana', 'foo': 2.35, 'bar': 'kiwi'},
            object._database_cache,
            "database cache synced"
        )

        self.assertEqual(
            2.35,
            object.foo,
            "check that object.foo is correct before the change"
        )

        object.foo = 69.42

        self.assertEqual(
            69.42,
            object.foo,
            "check that object.foo is correct after the change"
        )

        self.assertDictEqual(
            {'id_part_1': 12, 'id_part_2': 'banana', 'foo': 2.35, 'bar': 'kiwi'},
            object._database_cache,
            "database cache still unchanged after property value change"
        )

        self.assertDictEqual(
            {"foo": 69.42},
            object.get_changes(),
            "get_changes() identifies the change in the foo attribute correctly"
        )


        sql, *where_values = object.build_sql_update()
        # noinspection SqlResolve,SqlNoDataSourceInspection
        self.assertEqual(
            dedent("""
                UPDATE "cool_table_name_yoooo"
                 SET "foo" = $1
                 WHERE "id_part_1" = $2 AND "id_part_2" = $3
                ;
            """).strip(),
            sql,
            "the produced update SQL is correct"
        )

        self.assertEqual(
            [   # the value to change:
                69.42,
                # and the primary key(s):
                12, "banana",
            ],
            where_values,
            "make sure that the update's actual values are correct."
        )
    # end def
# end class


if __name__ == '__main__':
    unittest.main()
# end if



from fastorm import FastORM
from fastorm.classes import FieldInfo, FieldItem

from tests.tools_for_the_tests_of_fastorm import VerboseTestCase




class GetFieldsReferencesSimpleTest(VerboseTestCase):
    def test_simple_double_key_no_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id_part_1', 'id_part_2']

            id_part_1: int
            id_part_2: str
        # end class

        with self.subTest():
            refs = Test1.get_fields_references(recursive=False)
            expected = {
                'id_part_1': FieldInfo(True, [FieldItem('id_part_1', int)]),
                'id_part_2': FieldInfo(True, [FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs)
        # end with

        with self.subTest():
            refs = Test1.get_fields_references(recursive=False)
            expected = {
                'id_part_1': FieldInfo(True, [FieldItem('id_part_1', int)]),
                'id_part_2': FieldInfo(True, [FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs)
        # end with
    # end def

    def test_simple_single_key_no_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id']

            id: int
            text: str
        # end class

        with self.subTest():
            refs = Test1.get_fields_references(recursive=False)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'text': FieldInfo(False, [FieldItem('text', str)]),
            }
            self.assertEqual(expected, refs)
        # end with

        with self.subTest():
            refs = Test1.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'text': FieldInfo(False, [FieldItem('text', str)]),
            }
            self.assertEqual(expected, refs)
        # end with
    # end def

    def test_simple_double_key_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id_part_1', 'id_part_2']

            id_part_1: int
            id_part_2: str
        # end class
        with self.subTest():
            refs = Test1.get_fields_references(recursive=True)
            expected = {
                'id_part_1': FieldInfo(True, [FieldItem('id_part_1', int)]),
                'id_part_2': FieldInfo(True, [FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs)
        # end with

        with self.subTest():
            refs = Test1.get_fields_references(recursive=True)
            expected = {
                'id_part_1': FieldInfo(True, [FieldItem('id_part_1', int)]),
                'id_part_2': FieldInfo(True, [FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs)
        # end with
    # end def

    def test_simple_single_key_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id']

            id: int
            text: str
        # end class

        with self.subTest():
            refs = Test1.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'text': FieldInfo(False, [FieldItem('text', str)]),
            }
            self.assertEqual(expected, refs)
        # end with

        with self.subTest():
            refs = Test1.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'text': FieldInfo(False, [FieldItem('text', str)]),
            }
            self.assertEqual(expected, refs)
        # end with
    # end def
# end def


class GetFieldsReferencesSingleReferenceTest(VerboseTestCase):
    def test_single_reference_double_key_no_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id_part_1', 'id_part_2']

            id_part_1: int
            id_part_2: str
        # end class

        with self.subTest(msg='Test1, recursive=False'):
            refs = Test1.get_fields_references(recursive=False)
            expected = {
                'id_part_1': FieldInfo(True, [FieldItem('id_part_1', int)]),
                'id_part_2': FieldInfo(True, [FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs, msg='Test1, recursive=False')
        # end with

        with self.subTest(msg='Test1, recursive=True'):
            refs = Test1.get_fields_references(recursive=True)
            expected = {
                'id_part_1': FieldInfo(True, [FieldItem('id_part_1', int)]),
                'id_part_2': FieldInfo(True, [FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs, msg='Test1, recursive=True')
        # end with

        class Test2(FastORM):
            _table_name = 'test2'
            _primary_keys = ['id']

            id: int
            test_one: Test1
        # end class

        with self.subTest(msg='Test2, recursive=False'):
            refs = Test2.get_fields_references(recursive=False)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'test_one': FieldInfo(False, [FieldItem('test_one', Test1)]),
            }
            self.assertEqual(expected, refs, msg='recursive=False')
        # end with

        with self.subTest(msg='Test2, recursive=True'):
            refs = Test2.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'test_one__id_part_1': FieldInfo(False, [FieldItem('test_one', Test1), FieldItem('id_part_1', int)]),
                'test_one__id_part_2': FieldInfo(False, [FieldItem('test_one', Test1), FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs, msg='Test2, recursive=True')
        # end with
    # end def

    def test_single_reference_no_pk_single_key_no_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id']

            id: int
            text: str
        # end class

        with self.subTest(msg='Test1, recursive=False'):
            refs = Test1.get_fields_references(recursive=False)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'text': FieldInfo(False, [FieldItem('text', str)]),
            }
            self.assertEqual(expected, refs, msg='Test1, recursive=False')
        # end with

        with self.subTest(msg='Test1, recursive=True'):
            refs = Test1.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'text': FieldInfo(False, [FieldItem('text', str)]),
            }
            self.assertEqual(expected, refs, msg='Test1, recursive=True')
        # end with

        class Test2(FastORM):
            _table_name = 'test2'
            _primary_keys = ['id']

            id: int
            test_one: Test1
        # end class

        with self.subTest(msg='Test2, recursive=False'):
            refs = Test2.get_fields_references(recursive=False)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'test_one': FieldInfo(False, [FieldItem('test_one', Test1)]),
            }
            self.assertEqual(expected, refs, msg='Test2, recursive=False')
        # end with

        with self.subTest(msg='Test2, recursive=True'):
            refs = Test2.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'test_one__id': FieldInfo(False, [FieldItem('test_one', Test1), FieldItem('id', int)]),
            }
            self.assertEqual(expected, refs, msg='Test2, recursive=True')
        # end with
    # end def

    def test_single_reference_no_pk_double_key_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id_part_1', 'id_part_2']

            id_part_1: int
            id_part_2: str
        # end class

        class Test2(FastORM):
            _table_name = 'test2'
            _primary_keys = ['id']

            id: int
            test_one: Test1
        # end class

        with self.subTest(msg='recursive=False'):
            refs = Test2.get_fields_references(recursive=False)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'test_one': FieldInfo(False, [FieldItem('test_one', Test1)]),
            }
            self.assertEqual(expected, refs, msg='recursive=False')
        # end with

        with self.subTest(msg='recursive=True'):
            refs = Test2.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'test_one__id_part_1': FieldInfo(False, [FieldItem('test_one', Test1), FieldItem('id_part_1', int)]),
                'test_one__id_part_2': FieldInfo(False, [FieldItem('test_one', Test1), FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs)
        # end with
    # end def

    def test_single_reference_no_pk_single_key_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id']

            id: int
            text: str
        # end class

        class Test2(FastORM):
            _table_name = 'test2'
            _primary_keys = ['id']

            id: int
            test_one: Test1
        # end class

        with self.subTest(msg='recursive=False'):
            refs = Test2.get_fields_references(recursive=False)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'test_one': FieldInfo(False, [FieldItem('test_one', Test1)]),
            }
            self.assertEqual(expected, refs)
        # end with

        with self.subTest(msg='recursive=True'):
            refs = Test2.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(True, [FieldItem('id', int)]),
                'test_one__id': FieldInfo(False, [FieldItem('test_one', Test1), FieldItem('id', int)]),
            }
            self.assertEqual(expected, refs)
        # end with
    # end def

    def test_single_reference_stacked_pk_key_no_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id_part_1', 'id_part_2']

            id_part_1: int
            id_part_2: str
        # end class

        class Test2(FastORM):
            _table_name = 'test2'
            _primary_keys = ['test_one']

            id: int
            test_one: Test1
        # end class

        with self.subTest(msg='recursive=False'):
            refs = Test2.get_fields_references(recursive=False)
            expected = {
                'id': FieldInfo(False, [FieldItem('id', int)]),
                'test_one': FieldInfo(True, [FieldItem('test_one', Test1)]),
            }
            self.assertEqual(expected, refs)
        # end with

        with self.subTest(msg='recursive=True'):
            refs = Test2.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(False, [FieldItem('id', int)]),
                'test_one__id_part_1': FieldInfo(True, [FieldItem('test_one', Test1), FieldItem('id_part_1', int)]),
                'test_one__id_part_2': FieldInfo(True, [FieldItem('test_one', Test1), FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs)
        # end with
    # end def

    def test_single_reference_single_key_recursive(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id']

            id: int
            text: str
        # end class

        class Test2(FastORM):
            _table_name = 'test2'
            _primary_keys = ['test_one']

            id: int
            test_one: Test1
        # end class

        with self.subTest(msg='recursive=True'):
            refs = Test2.get_fields_references(recursive=False)
            expected = {
                'id': FieldInfo(False, [FieldItem('id', int)]),
                'test_one': FieldInfo(True, [FieldItem('test_one', Test1)]),
            }
            self.assertEqual(expected, refs)
        # end with

        with self.subTest(msg='recursive=True'):
            refs = Test2.get_fields_references(recursive=True)
            expected = {
                'id': FieldInfo(False, [FieldItem('id', int)]),
                'test_one__id': FieldInfo(True, [FieldItem('test_one', Test1), FieldItem('id', int)]),
            }
            self.assertEqual(expected, refs)
        # end with
    # end def
# end class


class GetFieldsReferencesMultiLayerReferenceTest(VerboseTestCase):
    def test_two_level_on_double_reference_1(self):
        class Test1(FastORM):
            _table_name = 'test1'
            _primary_keys = ['id_part_1', 'id_part_2']

            id_part_1: int
            id_part_2: str
        # end class

        with self.subTest(msg='Test1, recursive=False'):
            refs = Test1.get_fields_references(recursive=False)
            expected = {
                'id_part_1': FieldInfo(True, [FieldItem('id_part_1', int)]),
                'id_part_2': FieldInfo(True, [FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs, msg='Test1, recursive=False')
        # end with

        with self.subTest(msg='Test1, recursive=True'):
            refs = Test1.get_fields_references(recursive=True)
            expected = {
                'id_part_1': FieldInfo(True, [FieldItem('id_part_1', int)]),
                'id_part_2': FieldInfo(True, [FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs, msg='Test1, recursive=True')
        # end with

        class Test2(FastORM):
            _table_name = 'test2'
            _primary_keys = ['test_one']

            some_number: int
            test_one: Test1
        # end class

        with self.subTest(msg='Test2, recursive=False'):
            refs = Test2.get_fields_references(recursive=False)
            expected = {
                'some_number': FieldInfo(False, [FieldItem('some_number', int)]),
                'test_one': FieldInfo(True, [FieldItem('test_one', Test1)]),
            }
            self.assertEqual(expected, refs, msg='Test2, recursive=False')
        # end with

        with self.subTest(msg='Test2, recursive=True'):
            refs = Test2.get_fields_references(recursive=True)
            expected = {
                'some_number': FieldInfo(False, [FieldItem('some_number', int)]),
                'test_one__id_part_1': FieldInfo(True, [FieldItem('test_one', Test1), FieldItem('id_part_1', int)]),
                'test_one__id_part_2': FieldInfo(True, [FieldItem('test_one', Test1), FieldItem('id_part_2', str)]),
            }
            self.assertEqual(expected, refs, msg='Test2, recursive=True')
        # end with

        class Test3(FastORM):
            _table_name = 'test3'
            _primary_keys = ['test_two']

            test_two: Test2
            title: str
        # end class

        with self.subTest(msg='Test3, recursive=False'):
            refs = Test3.get_fields_references(recursive=False)
            expected = {
                'test_two': FieldInfo(True, [FieldItem('test_two', Test2)]),
                'title': FieldInfo(False, [FieldItem('title', str)]),
            }
            self.assertEqual(expected, refs, msg='Test3, recursive=False')
        # end with

        with self.subTest(msg='Test3, recursive=True'):
            refs = Test3.get_fields_references(recursive=True)
            expected = {
                'test_two__test_one__id_part_1': FieldInfo(True, [FieldItem('test_two', Test2), FieldItem('test_one', Test1), FieldItem('id_part_1', int)]),
                'test_two__test_one__id_part_2': FieldInfo(True, [FieldItem('test_two', Test2), FieldItem('test_one', Test1), FieldItem('id_part_2', str)]),
                'title': FieldInfo(False, [FieldItem('title', str)]),
            }
            self.assertEqual(expected, refs, msg='Test3, recursive=True')
        # end with
    # end def

    def test_two_level_on_double_reference_2(self):
        class Test1A(FastORM):
            _table_name = 'test1a'
            _primary_keys = ['id_part_1', 'id_part_2']

            id_part_1: int
            id_part_2: str
        # end class

        class Test1B(FastORM):
            _table_name = 'test2b'
            _primary_keys = ['id']

            id: int
        # end class

        class Test2(FastORM):
            _table_name = 'test2'
            _primary_keys = ['test_one_a', 'test_one_b']

            some_number: int
            test_one_a: Test1A
            test_one_b: Test1B
        # end class

        class Test3(FastORM):
            _table_name = 'test3'
            _primary_keys = ['test_two', 'test_one_b']

            test_two: Test2
            test_one_b: Test1B
            title: str
        # end class

        with self.subTest(msg='recursive=False'):
            refs = Test3.get_fields_references(recursive=False)
            expected = {
                'test_two': FieldInfo(True, [FieldItem('test_two', Test2)]),
                'test_one_b': FieldInfo(True, [FieldItem('test_one_b', Test1B)]),
                'title': FieldInfo(False, [FieldItem('title', str)]),
            }
            self.assertEqual(expected, refs, msg='recursive=False')
        # end with

        with self.subTest(msg='recursive=True'):
            refs = Test3.get_fields_references(recursive=True)
            expected = {
                'test_two__test_one_a__id_part_1': FieldInfo(True, [FieldItem('test_two', Test2), FieldItem('test_one_a', Test1A), FieldItem('id_part_1', int)]),
                'test_two__test_one_a__id_part_2': FieldInfo(True, [FieldItem('test_two', Test2), FieldItem('test_one_a', Test1A), FieldItem('id_part_2', str)]),
                'test_two__test_one_b__id': FieldInfo(True, [FieldItem('test_two', Test2), FieldItem('test_one_b', Test1B), FieldItem('id', int)]),
                'test_one_b__id': FieldInfo(is_primary_key=True, types=[FieldItem(field='test_one_b', type_=Test1B), FieldItem(field='id', type_=int)]),
                'title': FieldInfo(False, [FieldItem('title', str)]),
            }
            self.assertEqual(expected, refs, msg='recursive=True')
        # end with
    # end def
# end class

if __name__ == '__main__':
    unittest.main()



import unittest
from textwrap import dedent

from fastorm import FastORM


class OtherTable(FastORM):
    """
        CREATE TABLE "other_table" (
          "id_part_1" TEXT NOT NULL,
          "id_part_2" BIGINT NOT NULL,
          PRIMARY KEY ("id_part_1", "id_part_2")
        );
    """
    _table_name = 'cool_table_name_yoooo'
    _primary_keys = ['id_part_1', 'id_part_2']
    _ignored_fields = []

    id_part_1: int
    id_part_2: str
    foo: float
    bar: str
# end class


class UpdateRowTestCase(unittest.TestCase):
    def test_foo(self):
        object = OtherTable(
            id_part_1=12,
            id_part_2='banana',
            foo=2.35,
            bar="kiwi",
        )

        sql, *where_values = object.build_sql_delete()
        # noinspection SqlResolve,SqlNoDataSourceInspection
        self.assertEqual(
            dedent("""
                DELETE FROM "cool_table_name_yoooo"
                 WHERE "id_part_1" = $1 AND "id_part_2" = $2
                ;
            """).strip(),
            sql,
            "the produced update SQL is correct"
        )
    # end def
# end class


if __name__ == '__main__':
    unittest.main()
# end if



"""
A TestRunner for use with the Python unit testing framework. It
generates a tabular report to show the result at a glance.
The simplest way to use this is to invoke its main method. E.g.
    import unittest
    import TestRunner
    ... define your tests ...
    if __name__ == '__main__':
        TestRunner.main()
    # run the test
    runner.run(my_test_suite)
This TestRunner is based on HTMLTestRunner <http://tungwaiyip.info/software/HTMLTestRunner.html>
It is likely that I will rewrite this module form scracth soon.
By the way, HTMLTestRunner's license does not cover forking, given that I removed HTMLTestRunner's main characteristic(the HTML), I decided also removing the license. If I did not interpret the license properly, please, let me know.
HTMLTestRunner's author is Wai Yip Tung and I am grateful for his contribution.

Based on https://stackoverflow.com/a/31665827/3423324#display-python-unittest-results-in-nice-tabular-form
and https://gist.github.com/viniciusd/73e6eccd39dea5e714b1464e3c47e067
"""
import datetime

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import sys
import re
import time
import unittest


# ------------------------------------------------------------------------
# The redirectors below are used to capture output during testing. Output
# sent to sys.stdout and sys.stderr are automatically captured. However
# in some cases sys.stdout is already cached before HTMLTestRunner is
# invoked (e.g. calling logging.basicConfig). In order to capture those
# output, use the redirectors for the cached stream.
#
# e.g.
#   >>> logging.basicConfig(stream=HTMLTestRunner.stdout_redirector)
#   >>>

class OutputRedirector(object):
    """ Wrapper to redirect stdout or stderr """

    def __init__(self, fp):
        self.fp = fp

    def write(self, s):
        self.fp.write(s)

    def writelines(self, lines):
        self.fp.writelines(lines)

    def flush(self):
        self.fp.flush()


stdout_redirector = OutputRedirector(sys.stdout)
stderr_redirector = OutputRedirector(sys.stderr)


class Table(object):

    def __init__(self, padding='', allow_newlines=False):
        self.__columnSize__ = []
        self.__rows__ = []
        self.__titles__ = None
        self.padding = padding
        self.allow_newlines = allow_newlines

    @classmethod
    def remove_ansi_colors(cls, string):
        return re.sub("\033\[[0-9](?:;[0-9])?;[0-9]{1,2}m", "", string)
    # end def

    def __len__(self, x):
        return len(self.remove_ansi_colors(x))

    def addRow(self, row):
        rows = [[''] for l in range(len(row))]
        maxrows = 1
        for i, x in enumerate(row):
            for j, y in enumerate(x.split("\n")):
                if len(y) == 0 and self.allow_newlines == False:
                    continue
                try:
                    self.__columnSize__[i] = max(self.__columnSize__[i], self.__len__(y))
                except IndexError:
                    self.__columnSize__.append(self.__len__(y))
                rows[i].append(y)
                maxrows = max(j, maxrows)
        for i in range(len(rows)):
            rows[i] += (maxrows - (len(rows[i]) - 1)) * ['']
        for i in range(maxrows):
            self.__rows__.append([rows[j][i + 1] for j in range(len(row))])

    def addTitles(self, titles):
        for i, x in enumerate(titles):
            try:
                self.__columnSize__[i] = max(self.__columnSize__[i], self.__len__(x))
            except IndexError:
                self.__columnSize__.append(self.__len__(x))
        self.__titles__ = titles

    def __str__(self):
        """
        Markdown version
        :return:
        """
        lines = []
        if self.__titles__ is None:
            title = " | ".join(f'{".": ^ {chars}}' for chars in self.__columnSize__)
        else:
            if len(self.__titles__) < len(self.__columnSize__):
                self.__titles__ += ((len(self.__columnSize__) - len(self.__titles__)) * [''])
            for i, x in enumerate(self.__titles__):
                self.__titles__[i] = x.center(self.__columnSize__[i])
            title = " | ".join(self.__titles__)
        # end if
        lines.append(title)

        lines.append(" | ".join(chars * '-' for chars in self.__columnSize__))  # dash for every character and space on the outside.

        for x in self.__rows__:
            if len(x) < len(self.__columnSize__):
                x += ((len(self.__columnSize__) - len(x)) * [''])
            for i, c in enumerate(x):
                x[i] = self.remove_ansi_colors(c).ljust(self.__columnSize__[i]) + (len(c) - self.__len__(c) - 3) * ' '
            lines.append(" | ".join(x) + "")
        # end for

        lines.append("")

        return "\n".join(lines)
        # self.padding

    def __repr__(self):
        hline = self.padding + "+"
        for x in self.__columnSize__:
            hline += (x + 2) * '-' + '+'
        rows = []
        if self.__titles__ is None:
            title = ""
        else:
            if len(self.__titles__) < len(self.__columnSize__):
                self.__titles__ += ((len(self.__columnSize__) - len(self.__titles__)) * [''])
            for i, x in enumerate(self.__titles__):
                self.__titles__[i] = x.center(self.__columnSize__[i])
            title = self.padding + "| " + " | ".join(self.__titles__) + " |\n" + hline + "\n"
        for x in self.__rows__:
            if len(x) < len(self.__columnSize__):
                x += ((len(self.__columnSize__) - len(x)) * [''])
            for i, c in enumerate(x):
                x[i] = c.ljust(self.__columnSize__[i]) + (len(c) - self.__len__(c) - 3) * ' '
            rows.append(self.padding + "| " + " | ".join(x) + " |")
        return hline + "\n" + title + "\n".join(rows) + "\n" + hline + "\n"


class bcolors(object):
    FORMAT = {
        'Regular': '0',
        'Bold': '1',
        'Underline': '4',
        'High Intensity': '0',  # +60 on color
        'BoldHighIntensity': '1',  # +60 on color
    }
    START = "\033["
    COLOR = {
        'black': "0;30m",
        'red': "0;31m",
        'green': "0;32m",
        'yellow': "0;33m",
        'blue': "0;34m",
        'purple': "0;35m",
        'cyan': "0;36m",
        'white': "0;37m",
        'end': "0m",
    }

    def __getattr__(self, name):
        def handlerFunction(*args, **kwargs):
            return self.START + self.FORMAT['Regular'] + ";" + self.COLOR[name.lower()]

        return handlerFunction(name=name)
    # ----------------------------------------------------------------------


# Template

class Template_mixin(object):
    bc = bcolors()

    STATUS = {
        0: bc.GREEN + '✅ pass' + bc.END,
        1: bc.PURPLE + '❌ fail' + bc.END,
        2: bc.RED + '⚠️ error' + bc.END,
    }

    # ------------------------------------------------------------------------
    # Report
    #

    REPORT_TEST_WITH_OUTPUT_TMPL = r"""
   %(desc)s

        %(status)s
        %(script)s
"""  # variables: (tid, Class, style, desc, status)

    REPORT_TEST_NO_OUTPUT_TMPL = r"""
    %(desc)s
    %(status)s
"""  # variables: (tid, Class, style, desc, status)

    REPORT_TEST_OUTPUT_TMPL = r"""
%(output)s
"""  # variables: (id, output)


# -------------------- The end of the Template class -------------------


TestResult = unittest.TestResult


class _TestResult(TestResult):
    # note: _TestResult is a pure representation of results.
    # It lacks the output and reporting ability compares to unittest._TextTestResult.

    def __init__(self, verbosity=1):
        TestResult.__init__(self)
        self.stdout0 = None
        self.stderr0 = None
        self.success_count = 0
        self.failure_count = 0
        self.error_count = 0
        self.verbosity = verbosity

        # result is a list of result in 4 tuple
        # (
        #   result code (0: success; 1: fail; 2: error),
        #   TestCase object,
        #   Test output (byte string),
        #   stack trace,
        # )
        self.result = []

    def startTest(self, test):
        TestResult.startTest(self, test)
        # just one buffer for both stdout and stderr
        self.outputBuffer = StringIO()
        stdout_redirector.fp = self.outputBuffer
        stderr_redirector.fp = self.outputBuffer
        self.stdout0 = sys.stdout
        self.stderr0 = sys.stderr
        sys.stdout = stdout_redirector
        sys.stderr = stderr_redirector

    def complete_output(self):
        """
        Disconnect output redirection and return buffer.
        Safe to call multiple times.
        """
        if self.stdout0:
            sys.stdout = self.stdout0
            sys.stderr = self.stderr0
            self.stdout0 = None
            self.stderr0 = None
        return self.outputBuffer.getvalue()

    def stopTest(self, test):
        # Usually one of addSuccess, addError or addFailure would have been called.
        # But there are some path in unittest that would bypass this.
        # We must disconnect stdout in stopTest(), which is guaranteed to be called.
        self.complete_output()

    def addSuccess(self, test):
        self.success_count += 1
        TestResult.addSuccess(self, test)
        output = self.complete_output()
        self.result.append((0, test, output, ''))
        if self.verbosity > 1:
            sys.stderr.write('ok ')
            sys.stderr.write(str(test))
            sys.stderr.write('\n')
        else:
            pass  # sys.stderr.write('.')

    def addError(self, test, err):
        self.error_count += 1
        TestResult.addError(self, test, err)
        _, _exc_str = self.errors[-1]
        output = self.complete_output()
        self.result.append((2, test, output, _exc_str))
        if self.verbosity > 1:
            sys.stderr.write('E  ')
            sys.stderr.write(str(test))
            sys.stderr.write('\n')
        else:
            pass  # sys.stderr.write('E')

    def addFailure(self, test, err):
        self.failure_count += 1
        TestResult.addFailure(self, test, err)
        _, _exc_str = self.failures[-1]
        output = self.complete_output()
        self.result.append((1, test, output, _exc_str))
        if self.verbosity > 1:
            sys.stderr.write('F  ')
            sys.stderr.write(str(test))
            sys.stderr.write('\n')
        else:
            pass  # sys.stderr.write('F')


class TestRunner(Template_mixin):
    """
    """

    def __init__(self, stream=sys.stdout, verbosity=1, title=None, description=None):
        self.stream = stream
        self.verbosity = verbosity
        if title is None:
            self.title = 'Unit Test Report'
        else:
            self.title = title
        if description is None:
            self.description = ''
        else:
            self.description = description

        self.startTime = datetime.datetime.now()
        self.bc = bcolors()

    def run(self, test):
        "Run the given test case or test suite."
        result = _TestResult(self.verbosity)
        test(result)
        self.stopTime = datetime.datetime.now()
        self.generateReport(test, result)
        # print >> sys.stderr, '\nTime Elapsed: %s' % (self.stopTime-self.startTime)
        return result

    def sortResult(self, result_list):
        # unittest does not seems to run in any particular order.
        # Here at least we want to group them together by class.
        rmap = {}
        classes = []
        for n, test, output, error in result_list:
            testClass = test.__class__
            if testClass not in rmap:
                rmap[testClass] = []
                classes.append(testClass)
            rmap[testClass].append((n, test, output, error))
        r = [(testClass, rmap[testClass]) for testClass in classes]
        return r

    def getReportAttributes(self, result):
        """
        Return report attributes as a list of (name, value).
        Override this to add custom attributes.
        """
        startTime = str(self.startTime)[:19]
        duration = str(self.stopTime - self.startTime)
        status = []
        padding = 4 * ' '
        if result.success_count:
            status.append(padding + self.bc.GREEN + 'Pass:' + self.bc.END + ' %s\n' % result.success_count)
        if result.failure_count:
            status.append(padding + self.bc.PURPLE + 'Failure:' + self.bc.END + ' %s\n' % result.failure_count)
        if result.error_count:
            status.append(padding + self.bc.RED + 'Error:' + self.bc.END + ' %s\n' % result.error_count)
        if status:
            status = '\n' + ''.join(status)
        else:
            status = 'none'
        return [
            ('Start Time', startTime),
            ('Duration', duration),
            ('Status', status),
        ]

    def generateReport(self, test, result):
        report_attrs = self.getReportAttributes(result)
        heading = self._generate_heading(report_attrs)
        report = self._generate_report(result)
        output = self.title.rjust(30) + "\n" + \
                 heading + \
                 report
        try:
            self.stream.write(output.encode('utf8'))
        except TypeError:
            self.stream.write(output)

    def _generate_heading(self, report_attrs):
        a_lines = []
        for name, value in report_attrs:
            line = self.bc.CYAN + name + ": " + self.bc.END + value + "\n"
        a_lines.append(line)
        heading = ''.join(a_lines) + \
                  self.bc.CYAN + "Description:" + self.bc.END + self.description + "\n"
        return heading

    def _generate_report(self, result):
        rows = []
        sortedResult = self.sortResult(result.result)
        padding = 4 * ' '
        table = Table(padding=padding)
        table.addTitles(["Test group/Test case", "Count", "Pass", "Fail", "Error"])
        tests = ''
        for cid, (testClass, classResults) in enumerate(sortedResult):  # Iterate over the test cases
            classTable = Table(padding=2 * padding)
            classTable.addTitles(["Test name", "Status", "Stack"])
            # subtotal for a class
            np = nf = ne = 0
            for n, t, o, e in classResults:
                if n == 0:
                    np += 1
                elif n == 1:
                    nf += 1
                else:
                    ne += 1

            # format class description
            if testClass.__module__ == "__main__":
                name = testClass.__name__
            else:
                name = "%s.%s" % (testClass.__module__, testClass.__name__)
            tests += f'\n\n### `{name}` <a name="{name}" />\n\n'
            doc = testClass.__doc__ and testClass.__doc__.split("\n")[0] or ""
            desc = doc and '%s: %s' % (name, doc) or f'[`{name}`](#{name})'
            # style = ne > 0 and 'errorClass' or nf > 0 and 'failClass' or 'passClass',
            table.addRow([desc, str(np + nf + ne), str(np), str(nf), str(ne)])
            for tid, (n, test, output, error) in enumerate(classResults):  # Iterate over the unit tests
                t_name, t_stack, t_status = self._generate_report_test(cid, tid, n, test, output, error)
                classTable.addRow([t_name, t_status, t_stack])
            tests += str(classTable)
        table.addRow(
            ["Total", str(result.success_count + result.failure_count + result.error_count), str(result.success_count),
             str(result.failure_count), str(result.error_count)])
        report = self.bc.CYAN + "Summary: " + self.bc.END + "\n" + str(table) + tests
        return report

    def _generate_report_test(self, cid, tid, n, test, output, error):
        has_output = bool(output or error)
        tid = (n == 0 and 'p' or 'f') + 't%s.%s' % (cid + 1, tid + 1)
        name = test.id().split('.')[-1]
        doc = test.shortDescription() or ""
        desc = doc and ('%s: %s' % (name, doc)) or name
        tmpl = has_output and self.REPORT_TEST_WITH_OUTPUT_TMPL or self.REPORT_TEST_NO_OUTPUT_TMPL

        # o and e should be byte string because they are collected from stdout and stderr?
        if isinstance(output, str):
            # TODO: some problem with 'string_escape': it escape \n and mess up formating
            # uo = unicode(o.encode('string_escape'))
            try:
                uo = output.decode('latin-1')
            except AttributeError:
                uo = output
        else:
            uo = output
        if isinstance(error, str):
            # TODO: some problem with 'string_escape': it escape \n and mess up formating
            # ue = unicode(e.encode('string_escape'))
            try:
                ue = error.decode('latin-1')
            except AttributeError:
                ue = error
        else:
            ue = error

        script = self.REPORT_TEST_OUTPUT_TMPL % dict(
            output=uo + ue,
        )
        row = [desc, script, self.STATUS[n]]
        # row = tmpl % dict(
        #    tid = tid,
        #    desc = desc,
        #    script = script,
        #    status = self.STATUS[n],
        # )
        return row
        # if not has_output:
        #    return


##############################################################################
# Facilities for running tests from the command line
##############################################################################

# Note: Reuse unittest.TestProgram to launch test. In the future we may
# build our own launcher to support more specific command line
# parameters like test title, CSS, etc.
class TestProgram(unittest.TestProgram):
    """
    A variation of the unittest.TestProgram. Please refer to the base
    class for command line parameters.
    """

    def runTests(self):
        # Pick TestRunner as the default test runner.
        # base class's testRunner parameter is not useful because it means
        # we have to instantiate TestRunner before we know self.verbosity.
        if self.testRunner is None:
            self.testRunner = TestRunner(verbosity=self.verbosity)
        unittest.TestProgram.runTests(self)


main = TestProgram

##############################################################################
# Executing this module from the command line
##############################################################################

if __name__ == "__main__":
    main(module=None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Yes, this has a very long name, but we need to import it absolutely so better make sure it's unique.
"""
import contextlib
import unittest.case

from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if


def remove_prefix(line, prefix):
    try:
        return line.removeprefix(prefix)
    except AttributeError:
        if line.startswith(prefix):
            return line[len(prefix):]
        # end if
        return line
    # end try
# end def


def extract_sql_from_docstring(cls):
    return "\n".join(remove_prefix(line, '        ') for line in cls.__doc__.strip().splitlines() if not line.strip().startswith('#'))
# end def


# noinspection PyUnresolvedReferences,PyProtectedMember
_subtest_msg_sentinel = unittest.case._subtest_msg_sentinel


class VerboseTestCase(unittest.TestCase):
    show_real_diffs_in_pycharm_instead_of_having_subtests = True

    def subTest(self, msg=_subtest_msg_sentinel, **params):
        if not VerboseTestCase.show_real_diffs_in_pycharm_instead_of_having_subtests:
            return super().subTest(msg=msg, **params)
        else:
            @contextlib.contextmanager
            def subTestNoOP(msg=msg, **params):
                yield
            # end def
            return subTestNoOP(msg=msg, **params)
        # end if
    # end def

    def setUp(self) -> None:
        self.maxDiff = None
    # end def



import doctest
import fastorm
import unittest


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(module=fastorm))
    return tests
# end def


if __name__ == '__main__':
    unittest.main()
# end if



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""FastORM framework, easy to learn, fast to code"""
__author__ = 'luckydonald'
__version__ = "0.0.8"
__all__ = ['__author__', '__version__', 'FastORM', 'Autoincrement']

import ipaddress
import builtins
import datetime
import asyncpg
import decimal
import typing
import types
import uuid
import re
from typing import List, Dict, Any, Optional, Tuple, Type, get_type_hints, Union, TypeVar, Callable
from asyncpg import Connection, Pool
from luckydonaldUtils.exceptions import assert_type_or_raise
from luckydonaldUtils.logger import logging
from luckydonaldUtils.typing import JSONType

from pydantic import BaseModel
from pydantic.fields import ModelField, UndefinedType, Undefined, Field, PrivateAttr
from pydantic.typing import NoArgAnyCallable
from typeguard import check_type

from .classes import FieldInfo, FieldItem
from .compat import check_is_union_type, TYPEHINT_TYPE, check_is_generic_alias, check_is_annotated_type
from .compat import Annotated, NoneType

VERBOSE_SQL_LOG = True
CLS_TYPE = TypeVar("CLS_TYPE")


logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if


class FastORM(BaseModel):
    _table_name: str  # database table name we run queries against
    _ignored_fields: List[str]  # fields which never are intended for the database and will be excluded in every operation. (So are all fields starting with an underscore)
    _automatic_fields: List[str]  # fields the database fills in, so we will ignore them on INSERT.
    _primary_keys: List[str]  # this is how we identify ourself.
    _database_cache: Dict[str, JSONType] = PrivateAttr()  # stores the last known retrieval, so we can run UPDATES after you changed parameters.
    __selectable_fields: List[str] = PrivateAttr()  # cache for `cls.get_sql_fields()`

    def __init__(self, **data: Any):
        super().__init__(**data)
        self._database_cache: Dict[str, Any] = {}
    # end def

    def _database_cache_overwrite_with_current(self):
        """
        Resets the database cache from the current existing fields.
        This is used just after something is loaded from a database row.
        :return:
        """
        self._database_cache = {}
        for field in self.get_fields():
            self._database_cache[field] = getattr(self, field)
        # end if

    def _database_cache_remove(self):
        """
        Removes the database cache completely.
        This is used after deleting an entry.
        :return:
        """
        self._database_cache = {}
    # end def

    def as_dict(self) -> Dict[str, JSONType]:
        return self.dict()
    # end def

    @classmethod
    def get_fields_typehints(cls, *, flatten_table_references: bool = False) -> Dict[str, FieldInfo[ModelField]]:
        """
        Get's all fields which have type hints and thus we consider as fields for the database.
        Filters out constants (all upper case, like `CAPSLOCK_VARIABLE`) and hidden fields (starting with `_`).

        :param flatten_table_references:
                True if we should flatten the references to other table's primary key in the format of `f"{original_key}__{other_table_key}`.
                False to not resolve those fields, and instead return the type hint for the other FastORM class.
        :return: the dictionary with pydantic's ModelField descriptions.

        Example:

            >>> class OtherTable(FastORM):
            ...     _table_name = 'other_table'
            ...     _primary_keys = ['id_part_1', 'id_part_2']
            ...
            ...     id_part_1: int
            ...     id_part_2: str
            ... # end class
            ...

            >>> class ActualTable(FastORM):
            ...     _table_name = 'actual_table'
            ...     _primary_keys = ['cool_reference']
            ...     cool_reference: OtherTable
            ...

            >>> ActualTable.get_fields_typehints(flatten_table_references=False)
            {'cool_reference': FieldInfo(is_primary_key=True, types=[FieldItem(field='cool_reference', type_=ModelField(name='cool_reference', type=OtherTable, required=True))])}


            >>> ActualTable.get_fields_typehints(flatten_table_references=True)
            {'cool_reference__id_part_1': FieldInfo(is_primary_key=True, types=[FieldItem(field='cool_reference', type_=ModelField(name='cool_reference', type=OtherTable, required=True)), FieldItem(field='id_part_1', type_=ModelField(name='id_part_1', type=int, required=True))]), 'cool_reference__id_part_2': FieldInfo(is_primary_key=True, types=[FieldItem(field='cool_reference', type_=ModelField(name='cool_reference', type=OtherTable, required=True)), FieldItem(field='id_part_2', type_=ModelField(name='id_part_2', type=str, required=True))])}

        """
        _ignored_fields = cls.get_ignored_fields()
        references = cls.get_fields_references(recursive=flatten_table_references)

        # prepare the array with all the classes we wanna have typehints for.
        classes_typehints: Dict[Type, Dict[str, ModelField]] = {
            cls: {},
        }
        for long_key, field_reference in references.items():
            for interesting_type in field_reference.types[:-1]:  # the last one always is a native type (int, str, …), so not interesting
                classes_typehints[interesting_type.type_] = {}
            # end for
        # end for

        # now actually fill in type hints to that lookup table.
        for interesting_cls in classes_typehints.keys():
            interesting_cls: Type[FastORM]
            assert issubclass(interesting_cls, FastORM)
            type_hints: Dict[str, ModelField] = {
                key: value for key, value in interesting_cls.__fields__.items()
                if (
                    not key.startswith('_')
                    and not key.isupper()
                    and not key in _ignored_fields
                )
            }
            classes_typehints[interesting_cls] = type_hints
        # end for

        # now go through the references and apply the type hints
        result_hints: Dict[str, FieldInfo[ModelField]] = {}
        for long_key, field_reference in references.items():
            last_class = cls
            final_hint = FieldInfo(is_primary_key=field_reference.is_primary_key, types=[])
            for field_item in field_reference.types:
                type_hint = classes_typehints[last_class][field_item.field]
                final_hint.types.append(FieldItem(field=field_item.field, type_=type_hint))
                last_class = field_item.type_
            # end for
            result_hints[long_key] = final_hint
        # end for
        return result_hints
    # end def

    _GET_FIELDS_REFERENCES_TYPE = Dict[str, FieldInfo[typing.Union[type, typing.Type['FastORM']]]]

    @classmethod
    def get_fields_references(cls, *, recursive: bool = False) -> _GET_FIELDS_REFERENCES_TYPE:
        """
        Get's all fields which have type hints and thus we consider as fields for the database.
        Filters out constants (all upper case, like `CAPSLOCK_VARIABLE`) and hidden fields (starting with `_`).
        Then

        :param recursive: If we should not only take into account the current layer, but add the primary keys of the referenced tables to it as well, in the format `f"{our_key}__{referenced_primary_key}`.
        :return: the dictionary with the

        Example:

            >>> class OtherTable(FastORM):
            ...     _table_name = 'other_table'
            ...     _primary_keys = ['id_part_1', 'id_part_2']
            ...
            ...     id_part_1: int
            ...     id_part_2: str
            ... # end class
            ...

            >>> class ActualTable(FastORM):
            ...     _table_name = 'actual_table'
            ...     _primary_keys = ['cool_reference']
            ...     cool_reference: OtherTable
            ...     foobar: int
            ...

            >>> class ThirdTable(FastORM):
            ...     _table_name = 'third_table'
            ...     _primary_keys = ['id']
            ...     id: int
            ...     reference_to_other_table: OtherTable
            ...     reference_to_actual_table: ActualTable
            ...

            >>> class TableWithIdAndRefPK(FastORM):
            ...     _table_name = 'mayham_group_tables'
            ...     _primary_keys = ['id', 'reference']
            ...     _automatic_fields = ['id']
            ...     id: int
            ...     reference: OtherTable
            ...

            >>> class TableWithIdAndRefPK(FastORM):
            ...     _table_name = 'mayham_group_tables'
            ...     _primary_keys = ['id', 'reference']
            ...     _automatic_fields = ['id']
            ...     reference: OtherTable
            ...


            >>> ActualTable.get_fields_references(recursive=False)
            {'cool_reference': FieldInfo(is_primary_key=True, types=[FieldItem(field='cool_reference', type_=<class 'fastorm.OtherTable'>)]), 'foobar': FieldInfo(is_primary_key=False, types=[FieldItem(field='foobar', type_=<class 'int'>)])}

            >>> ThirdTable.get_fields_references(recursive=False)
            {'id': FieldInfo(is_primary_key=True, types=[FieldItem(field='id', type_=<class 'int'>)]), 'reference_to_other_table': FieldInfo(is_primary_key=False, types=[FieldItem(field='reference_to_other_table', type_=<class 'fastorm.OtherTable'>)]), 'reference_to_actual_table': FieldInfo(is_primary_key=False, types=[FieldItem(field='reference_to_actual_table', type_=<class 'fastorm.ActualTable'>)])}

            >>> ThirdTable.get_fields_references(recursive=True)
            {'id': FieldInfo(is_primary_key=True, types=[FieldItem(field='id', type_=<class 'int'>)]), 'reference_to_other_table__id_part_1': FieldInfo(is_primary_key=False, types=[FieldItem(field='reference_to_other_table', type_=<class 'fastorm.OtherTable'>), FieldItem(field='id_part_1', type_=<class 'int'>)]), 'reference_to_other_table__id_part_2': FieldInfo(is_primary_key=False, types=[FieldItem(field='reference_to_other_table', type_=<class 'fastorm.OtherTable'>), FieldItem(field='id_part_2', type_=<class 'str'>)]), 'reference_to_actual_table__cool_reference__id_part_1': FieldInfo(is_primary_key=False, types=[FieldItem(field='reference_to_actual_table', type_=<class 'fastorm.ActualTable'>), FieldItem(field='cool_reference', type_=<class 'fastorm.OtherTable'>), FieldItem(field='id_part_1', type_=<class 'int'>)]), 'reference_to_actual_table__cool_reference__id_part_2': FieldInfo(is_primary_key=False, types=[FieldItem(field='reference_to_actual_table', type_=<class 'fastorm.ActualTable'>), FieldItem(field='cool_reference', type_=<class 'fastorm.OtherTable'>), FieldItem(field='id_part_2', type_=<class 'str'>)])}

        """
        _ignored_fields = cls.get_ignored_fields()
        _primary_keys = cls.get_primary_keys_keys()
        # copy the type hints as we might add more type hints for the primary key fields of referenced models, and we wanna filter.
        type_hints = {
            key: value for key, value in cls.__fields__.items()
            if (
                not key.startswith('_')
                and not key.isupper()
                and not key in _ignored_fields
            )
        }
        return_val: FastORM._GET_FIELDS_REFERENCES_TYPE = {}
        for key, value in type_hints.items():
            type_hint = type_hints[key]
            inner_type = type_hint.type_
            other_class: Union[Type[FastORM], None] = None
            if (
                check_is_generic_alias(inner_type) and
                hasattr(inner_type, '__origin__') and
                type_hint.type_.__origin__ == typing.Union
            ):  # Union
                # it's a Union
                union_params = type_hint.type_.__args__[:]
                first_union_type = union_params[0]
                if issubclass(first_union_type, FastORM):
                    # we can have a reference to another Table, so it could be that
                    # the table ist the first entry and the actual field type is the second.
                    # Union[Table, int]
                    # Union[Table, Tuple[int, int]]
                    if not len(union_params) == 2:
                        raise TypeError(
                            f'Union with other table type must have it\'s primary key(s) as second argument: Union{union_params!r}'
                        )
                    # end if
                    implied_other_class_pk_types = union_params[1]
                    if (
                        check_is_generic_alias(implied_other_class_pk_types) and
                        hasattr(implied_other_class_pk_types, '__origin__') and
                        implied_other_class_pk_types.__origin__ == tuple and
                        hasattr(implied_other_class_pk_types, '__args__') and
                        implied_other_class_pk_types.__args__
                    ):
                        implied_other_class_pk_types = list(implied_other_class_pk_types.__args__)
                    else:
                        implied_other_class_pk_types = [implied_other_class_pk_types]
                    # end if
                    typehints = first_union_type.get_fields_typehints(flatten_table_references=True)
                    pk_keys_actual_types = [hint.resulting_type.type_ for hint in typehints.values() if hint.is_primary_key]
                    if implied_other_class_pk_types == pk_keys_actual_types:
                        # so basically the we know the referenced(!) `SomeTable` has `_id = ['id']`,
                        # and the referenced `SomeTable.id` is of type `int`,
                        # and now our given type, which is referencing it, is `Union[SomeTable, int]`, that is matching that.

                        # in other words:
                        # A `SomeTable` with primary key  being `int`        needs to have the reference called either `SomeTable` or `Union[SomeTable, int]` (`Union[SomeTable, [int]]` is possible too)
                        # A `SomeTable` with primary keys being `[int, str]` needs to have the reference called either `SomeTable` or `Union[SomeTable, [int, str]]`
                        other_class = first_union_type  # okey, we can use Table even if there's more. That way we have normalized it and there won't be an error later.
                    # end if
                # end if
            # end if
            if not other_class:
                try:
                    if issubclass(type_hint.type_, FastORM):
                        other_class: Type[FastORM] = type_hint.type_
                    # end if
                except TypeError:
                    pass
                # end try
            # end if
            other_class: Union[Type[FastORM], None]
            if not other_class:
                # is a regular key, just keep it as is
                # e.g. 'title': (False, [('title', str)]),
                return_val[key] = FieldInfo(key in _primary_keys, [FieldItem(key, type_hint.type_)])  # TODO: make a copy?
                # and then let's do the next key
                continue
            # end if

            # now we know: it's another FastORM table definition.
            assert issubclass(other_class, FastORM)
            # 'test_two__test_one_a__id_part_1': (True, [('test_two', Test2), ('test_one_a', Test1A), ('id_part_1', int)]),
            if not recursive:
                return_val[key] = FieldInfo(key in _primary_keys, [FieldItem(key, other_class)])  # TODO: make a copy?
                continue
            # end if
            other_refs = other_class.get_fields_references(recursive=True)

            for other_long_name, field_ref in other_refs.items():
                if not field_ref.is_primary_key:
                    continue
                # end if
                if not field_ref.types:
                    raise ValueError(f'Huh? No history at all! {other_long_name!r}, {other_class!r}, {field_ref.types!r}')
                # end if
                return_val[f'{key}__{other_long_name}'] = FieldInfo(
                        key in _primary_keys,
                        [FieldItem(key, other_class)] + field_ref.types
                )
                # end if
            # end for
        # end for
        return return_val
    # end def

    @classmethod
    def get_fields(cls, flatten_table_references: bool = False) -> List[str]:
        """
        Get's all fields which have type hints and thus we consider as fields for the database.
        Filters out constants (all upper case, like `CAPSLOCK_VARIABLE`) and hidden fields (starting with `_`).
        :return: a list with the keys.
        """
        return list(cls.get_fields_typehints(flatten_table_references=flatten_table_references).keys())
    # end def

    @classmethod
    def get_automatic_fields(cls) -> List[str]:
        _automatic_fields = getattr(cls, '_automatic_fields', [])[:]
        return _automatic_fields
    # end def

    @classmethod
    def get_ignored_fields(cls) -> List[str]:
        _ignored_fields = getattr(cls, '_ignored_fields', [])
        if isinstance(_ignored_fields, types.MemberDescriptorType):
            # basically that means it couldn't find any actually existing field
            _ignored_fields = []
        # end if
        _ignored_fields = [*_ignored_fields]  # make copy
        assert_type_or_raise(_ignored_fields, list, parameter_name=f'{cls.__name__}._ignored_fields')
        _ignored_fields += [
            '_table_name',
            '_ignored_fields',
            '_automatic_fields',
            '_primary_keys',
            '_database_cache',
            '__selectable_fields',
            f'_{cls.__name__!s}__selectable_fields',
            '__slots__'
        ]
        return _ignored_fields
    # end def

    @classmethod
    def get_sql_fields(cls) -> List[str]:
        key = f'_{cls.__name__!s}__selectable_fields'
        if getattr(cls, key, None) is None:
            setattr(cls, key, [field for field in cls.get_fields() if not field.startswith('_')])
        # end if
        return getattr(cls, key)
    # end if

    @classmethod
    def get_select_fields(cls, *, namespace=None) -> str:
        if namespace:
            return ', '.join([f'"{namespace}"."{field}" AS "{namespace} {field}"' for field in cls.get_sql_fields()])
        # end if
        return ', '.join([f'"{field}"' for field in cls.get_sql_fields()])
    # end def

    @classmethod
    def get_select_fields_len(cls) -> int:
        return len(cls.get_sql_fields())
    # end if

    @classmethod
    def get_table(cls) -> str:
        """
        Provides the name of the table in a already quoted way, ready to use in SQL queries.
        Note, that is naive quoting, and can be easily broken out of (possible SQL INJECTION), however the table name
        should never be user input anyway.

            >>> class Test(FastORM):
            ...   _table_name = 'sample'
            >>> Test.get_name()
            'sample'
            >>> Test.get_table()
            '"sample"'
            >>> print(Test.get_name())
            sample
            >>> print(Test.get_table())
            "sample"

        :return: The quoted table name
        """
        _table_name = cls.get_name()
        return f'"{_table_name}"'
    # end def

    @classmethod
    def get_name(cls) -> str:
        """
        The name of the table, as used with the database.
        :return: `table`
        """
        _table_name = getattr(cls, '_table_name')
        return _table_name
    # end def

    def build_sql_insert(
        self, *,
        ignore_setting_automatic_fields: Optional[bool] = None,
        on_conflict_upsert_field_list: Optional[List[str]] = None,
    ) -> Tuple[str, Any]:
        own_keys = self.get_fields()
        _ignored_fields = self.get_ignored_fields()
        _automatic_fields = self.get_automatic_fields()
        assert_type_or_raise(_ignored_fields, list, parameter_name='self._ignored_fields')
        assert_type_or_raise(_automatic_fields, list, parameter_name='self._automatic_fields')

        placeholder = []
        values: List[JSONType] = []
        keys = []
        upsert_fields = {}  # key is field name, values is the matching placeholder_index.
        placeholder_index = 0
        primary_key_index = 0
        for key in own_keys:
            if key in _ignored_fields:
                continue
            # end if
            is_automatic_field = None

            if ignore_setting_automatic_fields or on_conflict_upsert_field_list:
                is_automatic_field = key in _automatic_fields
            if ignore_setting_automatic_fields and is_automatic_field:
                continue
            # end if
            value = getattr(self, key)
            if ignore_setting_automatic_fields is None and value is None:
                continue
            # end if
            placeholder_index += 1
            placeholder.append(f'${placeholder_index}')
            if isinstance(value, dict):
                pass
                # value = json.dumps(value)
            # end if
            if isinstance(value, FastORM):
                # we have a different table in this table, so we probably want to go for it's `id` or whatever the primary key is.
                # if you got more than one of those PKs, simply specify them twice for both fields.
                value = value.get_primary_keys_values()[primary_key_index]
                primary_key_index += 1
            # end if

            values.append(value)
            keys.append(f'"{key}"')

            if on_conflict_upsert_field_list and not is_automatic_field:
                upsert_fields[key] = placeholder_index
            # end if
        # end if

        # noinspection SqlNoDataSourceInspection,SqlResolve
        sql = f'INSERT INTO {self.get_table()} ({",".join(keys)})\n VALUES ({",".join(placeholder)})'
        if on_conflict_upsert_field_list and upsert_fields:
            upsert_sql = ', '.join([f'"{key}" = ${placeholder_index}' for key, placeholder_index in upsert_fields.items()])
            upsert_fields_sql = ', '.join([f'"{field}"' for field in on_conflict_upsert_field_list])
            sql += f'\n ON CONFLICT ({upsert_fields_sql}) DO UPDATE SET {upsert_sql}'
        # end if
        if _automatic_fields:
            automatic_fields_sql = ', '.join([f'"{key}"' for key in _automatic_fields])
            sql += f'\n RETURNING {automatic_fields_sql}'
        # end if
        sql += '\n;'
        # noinspection PyRedundantParentheses
        return (sql, *values)
    # end def

    @classmethod
    async def get(cls: Type[CLS_TYPE], conn: Connection, **kwargs) -> Optional[CLS_TYPE]:
        """
        Retrieves a single Database element. Error if there are more matching ones.
        Like `.select(…)` but returns `None` for no matches, the match itself or an error if it's more than one row.

        :param conn:
        :param kwargs:
        :return:
        """
        rows = await cls.select(conn=conn, **kwargs)
        if len(rows) == 0:
            return None
        # end if
        assert len(rows) <= 1
        return rows[0]
    # end def

    @classmethod
    async def select(cls: Type[CLS_TYPE], conn: Connection, **kwargs) -> List[CLS_TYPE]:
        """
        Get's multiple ones.
        :param conn:
        :param kwargs:
        :return:
        """
        fetch_params = cls.build_sql_select(**kwargs)
        logger.debug(f'SELECT query for {cls.__name__}: {fetch_params[0]!r} with values {fetch_params[1:]}')
        rows = await conn.fetch(*fetch_params)
        return [cls.from_row(row) for row in rows]
    # end def

    @classmethod
    def build_sql_select(cls, **kwargs):
        _ignored_fields = cls.get_ignored_fields()
        typehints: Dict[str, Any] = get_type_hints(cls)
        non_ignored_fields = [field for field in cls.get_fields() if field not in _ignored_fields]
        fields = ','.join([
            f'"{field}"'
            for field in non_ignored_fields
            if not field.startswith('_')
        ])
        where_index = 0
        where_parts = []
        where_values = []
        # noinspection PyUnusedLocal
        where_wolf = None
        for key, value in kwargs.items():
            if key not in non_ignored_fields:
                raise ValueError(f'key {key!r} is not a non-ignored field!')
            # end if
            assert not isinstance(value, FastORM)
            # if isinstance(value, HelpfulDataclassDatabaseMixin):
            #     # we have a different table in this table, so we probably want to go for it's `id` or whatever the primary key is.
            #     # if you got more than one of those PKs, simply specify them twice for both fields.
            #     value = value.get_primary_keys_values()[primary_key_index]
            #     primary_key_index += 1
            # # end if
            where_index += 1
            is_in_list_clause = cls._param_is_list_of_multiple_values(key, value, typehints[key])
            if is_in_list_clause:
                assert isinstance(value, (list, tuple))
                assert len(value) >= 1
                if len(value) == 1:
                    # single element list -> normal where is fine -> so we go that route with it.
                    value = value[0]
                    is_in_list_clause = False
                # end if
            # end if
            if not is_in_list_clause:  # no else-if as is_in_list_clause could be set False again.
                where_parts.append(f'"{key}" = ${where_index}')
                where_values.append(value)
            else:  # is_in_list_clause is True
                where_part = ''
                for actual_value in value:
                    where_part += f'${where_index},'
                    where_values.append(actual_value)
                    where_index += 1
                # end for
                where_index -= 1  # we end up with incrementing once too much
                where_parts.append(f'"{key}" IN ({where_part.rstrip(",")})')
            # end if
        # end if
        where_sql = "" if not where_parts else f' WHERE {" AND ".join(where_parts)}'

        # noinspection SqlResolve,SqlNoDataSourceInspection
        sql = f'SELECT {fields} FROM "{cls._table_name}"{where_sql}'
        # noinspection PyRedundantParentheses
        return (sql, *where_values)
    # end def

    async def insert(
        self, conn: Connection, *,
        ignore_setting_automatic_fields: Optional[bool] = None,
        on_conflict_upsert_field_list: Optional[List[str]] = None,
        write_back_automatic_fields: bool = True,
    ):
        """
        :param conn: Database connection to run at.
        :param ignore_setting_automatic_fields:
            Skip setting fields marked as automatic, even if you provided.
            For example if the id field is marked automatic, as it's an autoincrement int.
            If `True`, setting `id=123` (commonly `id=None`) would be ignored, and instead the database assigns that value.
            If `False`, the value there will be written to the database.
            If `None`, it will be ignored as long as the value actually is None, but set if it is non-None.
            The default setting is `None`.
        :param on_conflict_upsert_field_list: List of fields which are expected to cause an duplicate conflict, and thus will instead be overwritten.
        :param write_back_automatic_fields: Apply the automatic fields back to this object.
                                            Ignored if `ignore_setting_automatic_fields` is False.
        :return:
        """
        fetch_params = self.build_sql_insert(
            ignore_setting_automatic_fields=ignore_setting_automatic_fields,
            on_conflict_upsert_field_list=on_conflict_upsert_field_list,
        )
        self._database_cache_overwrite_with_current()
        _automatic_fields = self.get_automatic_fields()
        if VERBOSE_SQL_LOG:
            fetch_params_debug = "\n".join([f"${i}={param!r}" for i, param in enumerate(fetch_params)][1:])
            logger.debug(f'INSERT query for {self.__class__.__name__}\nQuery:\n{fetch_params[0]}\nParams:\n{fetch_params_debug!s}')
        else:
            logger.debug(f'INSERT query for {self.__class__.__name__}: {fetch_params!r}')
        # end if
        updated_automatic_values_rows = await conn.fetch(*fetch_params)
        logger.debug(f'INSERT for {self.__class__.__name__}: {updated_automatic_values_rows} for {self}')
        assert len(updated_automatic_values_rows) == 1
        updated_automatic_values = updated_automatic_values_rows[0]
        if ignore_setting_automatic_fields and write_back_automatic_fields:
            for field in _automatic_fields:
                assert field in updated_automatic_values
                setattr(self, field, updated_automatic_values[field])
                self._database_cache[field] = updated_automatic_values[field]
            # end for
        # end if
    # end def

    def build_sql_update(self):
        """
        Builds a prepared SQL statement for update.
        Only fields with changed values will be updated in the database.
        However this one doesn't resets the cache for those, see `FastORM.update(…)` for that.

        :return: The SQL string followed by positional parameters for the `conn.execute(…)` method.
        """
        _table_name = getattr(self, '_table_name')
        _primary_keys = getattr(self, '_primary_keys')
        _database_cache = getattr(self, '_database_cache')
        _automatic_fields = self.get_automatic_fields()
        assert_type_or_raise(_table_name, str, parameter_name='self._table_name')
        assert_type_or_raise(_primary_keys, list, parameter_name='self._primary_keys')
        assert_type_or_raise(_database_cache, dict, parameter_name='self._database_cache')
        assert_type_or_raise(_automatic_fields, list, parameter_name='self._automatic_fields')

        # SET ...
        update_values = self.get_changes()

        # UPDATE ... SET ... WHERE ...
        placeholder_index = 0
        other_primary_key_index = 0
        values: List[Any] = []
        update_keys: List[str] = []  # "foo" = $1
        for key, value in update_values.items():
            value = getattr(self, key)
            placeholder_index += 1
            if isinstance(value, FastORM):
                # we have a different table in this table, so we probably want to go for it's `id` or whatever the primary key is.
                # if you got more than one of those PKs, simply specify them twice for both fields.
                value = value.get_primary_keys_values()[other_primary_key_index]
                other_primary_key_index += 1
            # end if

            values.append(value)
            update_keys.append(f'"{key}" = ${placeholder_index}')
        # end if

        # WHERE pk...
        primary_key_where: List[str] = []  # "foo" = $1
        for primary_key in _primary_keys:
            if primary_key in _database_cache:
                value = _database_cache[primary_key]
            else:
                value = getattr(self, primary_key)
            # end if
            placeholder_index += 1
            primary_key_where.append(f'"{primary_key}" = ${placeholder_index}')
            values.append(value)
        # end if
        logger.debug(f'Fields to UPDATE for selector {primary_key_where!r}: {update_values!r}')

        assert update_keys
        sql = f'UPDATE "{_table_name}"\n'
        sql += f' SET {",".join(update_keys)}\n'
        sql += f' WHERE {" AND ".join(primary_key_where)}\n'
        sql += ';'
        # noinspection PyRedundantParentheses
        return (sql, *values)
    # end def

    def get_changes(self) -> Dict:
        """
        Returns all values which got changed and are now different to the last downloaded database version.
        """
        own_keys = self.get_fields()
        _database_cache = self._database_cache
        _ignored_fields = self.get_ignored_fields()
        assert_type_or_raise(own_keys, list, parameter_name='own_keys')
        assert_type_or_raise(_database_cache, dict, parameter_name='self._database_cache')
        assert_type_or_raise(_ignored_fields, list, parameter_name='self._ignored_fields')

        update_values: Dict[str, Any] = {}
        for key in own_keys:
            if key.startswith('_') or key in _ignored_fields:
                continue
            # end if
            value = getattr(self, key)
            if key not in _database_cache:
                update_values[key] = value
            elif _database_cache[key] != value:
                update_values[key] = value
            # end if
        # end if
        return update_values
    # end def

    def has_changes(self) -> bool:
        """
        :returns: if we have unsaved changes.
        """
        return bool(self.get_changes())
    # end if

    async def update(self, conn: Connection) -> None:
        """
        Update the made changes to the database.
        Only fields with changed values will be updated in the database.

        :uses: FastORM.build_sql_update()
        """
        if not getattr(self, '_database_cache', None):
            return  # nothing to do.
        # end if
        fetch_params = self.build_sql_update()
        logger.debug(f'UPDATE query for {self.__class__.__name__}: {fetch_params!r}')
        update_status = await conn.execute(*fetch_params)
        logger.debug(f'UPDATE for {self.__class__.__name__}: {update_status} for {self}')
        self._database_cache_overwrite_with_current()
    # end if

    def build_sql_delete(self):
        _primary_keys = self.get_primary_keys_keys()
        _ignored_fields = self.get_ignored_fields()
        _database_cache = self._database_cache
        assert_type_or_raise(_ignored_fields, list, parameter_name='self._ignored_fields')
        assert_type_or_raise(_primary_keys, list, parameter_name='self._primary_keys')
        assert_type_or_raise(_database_cache, dict, parameter_name='self._database_cache')

        # DELETE FROM "name" WHERE pk...
        where_values = []
        placeholder_index = 0
        primary_key_parts: List[str] = []  # "foo" = $1
        for primary_key in _primary_keys:
            if primary_key in _database_cache:
                value = _database_cache[primary_key]
            else:
                value = getattr(self, primary_key)
            # end if
            placeholder_index += 1
            primary_key_parts.append(f'"{primary_key}" = ${placeholder_index}')
            where_values.append(value)
        # end if
        logger.debug(f'Fields to DELETE for selector {primary_key_parts!r}: {where_values!r}')

        # noinspection SqlWithoutWhere,SqlResolve,SqlNoDataSourceInspection
        sql = f'DELETE FROM {self.get_table()}\n'
        sql += f' WHERE {" AND ".join(primary_key_parts)}'
        sql += '\n;'
        # noinspection PyRedundantParentheses
        return (sql, *where_values)
    # end def

    async def delete(self, conn: Connection):
        fetch_params = self.build_sql_delete()
        logger.debug(f'DELETE query for {self.__class__.__name__}: {fetch_params!r}')
        delete_status = await conn.execute(*fetch_params)
        logger.debug(f'DELETE for {self.__class__.__name__}: {delete_status} for {self}')
        self._database_cache_remove()
    # end if

    def clone(self: CLS_TYPE) -> CLS_TYPE:
        return self.__class__(**self.as_dict())
    # end if

    @classmethod
    def get_primary_keys_keys(cls) -> List[str]:
        return cls._primary_keys
    # end def

    def get_primary_keys(self) -> Dict[str, Any]:
        _primary_keys = self.get_primary_keys_keys()
        return {k: v for k, v in self.as_dict().items() if k in _primary_keys}
    # end def

    def get_primary_keys_values(self):
        return list(self.get_primary_keys().values())
    # end def

    @classmethod
    def from_row(cls, row):
        # noinspection PyArgumentList
        instance = cls(*row)
        instance._database_cache_overwrite_with_current()
        return instance
    # end def

    _COLUMN_AUTO_TYPES: Dict[type, str] = {
        int: "BIGSERIAL",
    }

    _COLUMN_TYPES: Dict[type, str] = {
        bool: "BOOLEAN",
        bytes: "BYTEA",
        bytearray: "BYTEA",
        str: "TEXT",
        # Python Type
        # PostgreSQL Type
        # Source: https://magicstack.github.io/asyncpg/current/usage.html#type-conversion

        # anyenum
        # str

        # anyrange
        # asyncpg.Range

        # record
        # asyncpg.Record, tuple, Mapping

        # bit, varbit
        # asyncpg.BitString

        asyncpg.Box: "BOX",

        # cidr
        # ipaddress.IPv4Network, ipaddress.IPv6Network
        ipaddress.IPv4Network: "CIDR",
        ipaddress.IPv6Network: "CIDR",

        # inet
        # ipaddress.IPv4Interface, ipaddress.IPv6Interface, ipaddress.IPv4Address, ipaddress.IPv6Address
        ipaddress.IPv4Interface: "INET",
        ipaddress.IPv6Interface: "INET",
        ipaddress.IPv4Address: "INET",
        ipaddress.IPv6Address: "INET",

        # macaddr
        # str

        # time
        # offset-naïve datetime.time

        # time with time zone
        # offset-aware datetime.time
        datetime.time: "TIME",

        # timestamp
        # offset-naïve datetime.datetime
        #  +
        # timestamp with time zone
        # offset-aware datetime.datetime
        datetime.datetime: "TIMESTAMP",

        datetime.date: "DATE",  # must come after datetime.datetime as datetime is a subclass of this

        datetime.timedelta: "INTERVAL",

        # float, double precision
        # float [2]
        # Inexact single-precision float values may have a different representation when decoded into a Python float. This is inherent to the implementation of limited-precision floating point types. If you need the decimal representation to match, cast the expression to double or numeric in your query.
        float: "DOUBLE PRECISION",

        # smallint, integer, bigint
        # int
        int: "BIGINT",

        # numeric
        # Decimal
        decimal.Decimal: "NUMERIC",

        # json, jsonb
        # str
        dict: "JSONB",
        list: "JSONB",  # the special cases like INT[] will be processed beforehand.

        # line
        # asyncpg.Line
        asyncpg.Line: "LINE",

        # lseg
        # asyncpg.LineSegment
        asyncpg.LineSegment: "LSEG",

        # money
        # str

        asyncpg.Circle: "CIRCLE",

        # point
        # asyncpg.Point
        asyncpg.Point: "POINT",

        # polygon
        # asyncpg.Polygon
        asyncpg.Polygon: "POLYGON",

        # path
        # asyncpg.Path
        asyncpg.Path: "PATH",

        # uuid
        # uuid.UUID
        uuid.UUID: "UUID",

        BaseModel: "JSONB",
    }

    _COLUMN_TYPES_SPECIAL: Dict[Callable[[type], bool], str] = {
        lambda cls: hasattr(cls, 'to_dict'): _COLUMN_TYPES[dict],
        lambda cls: hasattr(cls, 'to_array'): _COLUMN_TYPES[dict],  # pytgbot object uses to_array
    }

    _COLUMN_AUTO_TYPES_SPECIAL: Dict[Callable[[type], bool], str] = {
    }

    @classmethod
    def _match_type(cls, python_type: type, *, automatic: bool) -> str:
        try:
            issubclass(python_type, object)
        except TypeError:  # issubclass() arg 1 must be a class
            raise TypeError(f'Could not process type {python_type} as a python type. Probably a typing annotation?.')
        if automatic:
            for sql_py_type, sql_type in cls._COLUMN_AUTO_TYPES.items():
                if issubclass(python_type, sql_py_type):
                    return sql_type
                # end if
            # end for
            for check_function, sql_type in cls._COLUMN_AUTO_TYPES_SPECIAL.items():
                if check_function(python_type):
                    return sql_type
                # end if
            # end for
        # end if
        for sql_py_type, sql_type in cls._COLUMN_TYPES.items():
            if issubclass(python_type, sql_py_type):
                return sql_type
            # end if
        # end for
        for check_function, sql_type in cls._COLUMN_TYPES_SPECIAL.items():
            if check_function(python_type):
                return sql_type
            # end if
        # end for
        raise TypeError(f'Could not process type {python_type} as database type.')
    # end def

    @classmethod
    async def create_table(
        cls,
        conn: Connection,
        if_not_exists: bool = False,
        psycopg2_conn: Union['psycopg2.extensions.connection', 'psycopg2.extensions.cursor', None] = None,
    ):
        """
        Builds and executes a CREATE TABLE statement.

        :param conn: the `asyncpg` database connection to execute this with.

        :param if_not_exists:
            If the table definition should include IF NOT EXISTS, thus not producing an error if it does, but instead being silently ignored.

        :param psycopg2_conn:
            If you have complex default types for your fields (everything other than None, bool, int, and pure ascii strings),
            The psycopg2 library is used to build a injection safe SQL string.
            Therefore then psycopg2 has to be installed (pip install psycopg2-binary),
            and a connection (or cursor) to the database must be provided (psycopg2_conn = psycopg2.connect(…)).
        :return:
        """
        create_params = cls.build_sql_create(if_not_exists=if_not_exists, psycopg2_conn=psycopg2_conn if psycopg2_conn else conn)
        logger.debug(f'CREATE query for {cls.__name__}: {create_params!r}')
        crate_status = await conn.execute(*create_params)
        logger.debug(f'CREATEed {cls.__name__}: {crate_status}')
    # end if

    @classmethod
    def build_sql_create(
        cls,
        if_not_exists: bool = False,
        psycopg2_conn: Union['psycopg2.extensions.connection', 'psycopg2.extensions.cursor', None] = None,
    ) -> Tuple[str, Any]:
        """
        Builds a CREATE TABLE statement.

        :param if_not_exists:
            If the table definition should include IF NOT EXISTS, thus not producing an error if it does, but instead being silently ignored.

        :param psycopg2_conn:
            If you have complex default types for your fields (everything other than None, bool, int, and pure ascii strings),
            The psycopg2 library is used to build a injection safe SQL string.
            Therefore then psycopg2 has to be installed (pip install psycopg2-binary),
            and a connection to the database must be provided.
            This can be either a psycopg2 connection (psycopg2_conn = psycopg2.connect(…)),
            a psycopg2 cursor.
            If a asyncpg Connection is given we try to create a psycopg2 connection from it automatically.
        :return:
        """
        assert issubclass(cls, BaseModel)  # because we no longer use typing.get_type_hints, but pydantic's `cls.__fields__`
        _table_name = getattr(cls, '_table_name')
        _automatic_fields = cls.get_automatic_fields()
        assert_type_or_raise(_table_name, str, parameter_name='cls._table_name')
        assert_type_or_raise(_automatic_fields, list, parameter_name='cls._automatic_fields')
        _ignored_fields = cls.get_ignored_fields()

        type_hints = cls.get_fields_typehints(flatten_table_references=True)
        primary_keys = [key for key, field_typehint in type_hints.items() if field_typehint.is_primary_key]
        single_primary_key = len(primary_keys) == 1

        # .required tells us if we have a default value set or not.
        # .allow_none tells us if None is supported
        # .default tells us what default (or None)

        placeholder_index = 0
        placeholder_values = []
        type_definitions = []
        all_defaults_are_simple_types_and_save_to_concatenate = True  # they only contain ints, boolean and None

        for key, type_hint_info in type_hints.items():
            type_hint = type_hint_info.types[-1].type_
            is_automatic_field = key in _automatic_fields

            is_optional, sql_type = cls.match_type(type_hint=type_hint, is_automatic_field=is_automatic_field, key=key)
            if is_automatic_field:
                is_optional = False
            # end if

            # Now let's build that column's sql part

            # column_name, data_type:
            type_definition_parts = [f'  "{key}"', sql_type]

            # column_constraints:
            if not is_optional:
                type_definition_parts.append("NOT NULL")
            # end if

            # has it a default value?
            if not isinstance(type_hint.field_info.default, UndefinedType):
                type_definition_parts.append(f'DEFAULT {{default_placeholder_{placeholder_index}}}')
                placeholder_index += 1
                default_value = type_hint.field_info.default
                if not (
                    default_value is None or
                    isinstance(default_value, (int, bool)) or
                    (isinstance(default_value, str) and default_value.isascii())  # py 3.7+, see https://stackoverflow.com/a/51141941/3423324#how-to-check-if-a-string-in-python-is-in-ascii
                ):
                    all_defaults_are_simple_types_and_save_to_concatenate = False
                # end if
                placeholder_values.append(default_value)
            # end if
            if single_primary_key and key in primary_keys:
                type_definition_parts.append('PRIMARY KEY')
            # end if

            # merge the line
            type_definitions.append(" ".join(type_definition_parts))
        # end for
        sql_lines = [
            f"CREATE TABLE {'IF NOT EXISTS ' if if_not_exists else ''}{cls.get_table()} (",
            ",\n".join(type_definitions),
        ]
        if len(primary_keys) > 1:
            sql_lines[-1] += ','
            # keys = [key for key, item in type_hints.items()]
            sql_lines.append(f'''  PRIMARY KEY ({', '.join(f'"{key}"' for key in primary_keys)})''')
        # end if
        sql_lines.append(");")

        # references to other tables
        references_types: Dict[str, FieldInfo[ModelField]] = {
            # everything with a different table will have more than one type in there.
            long_key: field_typehint for long_key, field_typehint in cls.get_fields_references(recursive=True).items() if len(field_typehint.types) > 1
        }

        if references_types:
            index_lines = []
            reference_lines = []
            current_table = cls.get_name()
            for current_table_field, field_typehint in references_types.items():
                field_typehint: FieldInfo[ModelField]
                assert issubclass(field_typehint.referenced_type, FastORM)
                referenced_table = field_typehint.referenced_type.get_name()
                referenced_table_field = field_typehint.referenced_field
                # noinspection SqlNoDataSourceInspection,SqlResolve
                index_lines.append(f'CREATE INDEX "idx_{current_table}___{current_table_field}" ON "{current_table}" ("{current_table_field}");')
                # noinspection SqlNoDataSourceInspection,SqlResolve
                reference_lines.append(f'ALTER TABLE "{current_table}" ADD CONSTRAINT "fk_{current_table}___{current_table_field}" FOREIGN KEY ("{current_table_field}") REFERENCES "{referenced_table}" ("{referenced_table_field}") ON DELETE CASCADE;')
            # end for
            sql_lines.extend(index_lines)
            sql_lines.extend(reference_lines)
        # end if

        sql = "\n".join(
            sql_lines
        )
        if placeholder_values:
            # we do have at least one row like
            # "blah" SOME_TYPE ... DEFAULT {default_placeholder_0}
            # or maybe even more.
            # now we need to fill those in.
            # To do it safely, we use `psycopg2.sql.SQL` and `psycopg2.sql.Literal` as soon as we have complex types.
            if all_defaults_are_simple_types_and_save_to_concatenate:
                # special case where it's enough to use python only formatting, as we have only NONE, bool, ints and pure ascii strings.
                formatting_dict = {}
                for i, default_value in enumerate(placeholder_values):
                    assert_type_or_raise(default_value, None, bool, int, str, parameter_name=f'default_values[{i}]')
                    if default_value is None:
                        sql_value = 'NULL'
                    elif isinstance(default_value, bool):
                        sql_value = 'true' if default_value else 'false'
                    elif isinstance(default_value, int):
                        sql_value = str(default_value)
                    else:  # str
                        assert isinstance(default_value, str)
                        assert default_value.isascii()
                        sql_value = "'" + default_value.replace("'", "''") + "'"
                    # end if
                    formatting_dict[f'default_placeholder_{i}'] = sql_value
                # end if
                sql = sql.format(**formatting_dict)
            else:
                try:
                    from psycopg2 import sql as sql_escaping  # you need to have `psycopg2-binary` installed.
                    from psycopg2 import extensions as ext, connect as psycopg2_connect
                except (ImportError, ModuleNotFoundError):
                    # enhance error message with useful information
                    raise ImportError(
                        'For using complex default values (everything other than None, bool, int, and pure ascii strings) '
                        'psycopg2 needs to be installed (pip install psycopg2-binary).'
                    )
                # end try
                try:
                    assert_type_or_raise(psycopg2_conn, Connection, ext.connection, ext.cursor, parameter_name='psycopg2_conn')
                except TypeError:
                    # enhance error message with useful information
                    error_class = ValueError if psycopg2_conn is None else TypeError
                    raise error_class(
                        'For using complex default values (everything other than None, bool, int, and pure ascii strings) '
                        'a psycopg2 connection or cursor needs to be provided as the psycopg2_conn parameter.'
                    )
                # end try
                if isinstance(psycopg2_conn, Connection):
                    # if it's a asyncpg Connection, try to build the needed psycopg2 connection from it.
                    # noinspection PyProtectedMember
                    params = dict(
                        database=psycopg2_conn._params.database,  # aka dbname
                        user=psycopg2_conn._params.user,
                        password=psycopg2_conn._params.password,
                        host=psycopg2_conn._addr[0] if isinstance(psycopg2_conn._addr, tuple) else psycopg2_conn._addr,
                        port=psycopg2_conn._addr[1] if isinstance(psycopg2_conn._addr, tuple) else None,
                    )
                    unix_socket_match = re.match(r'^(?P<path>/.*)(?<=/)\.s\.PGSQL\.(?P<port>\d+)$', params['host'])  # https://regex101.com/r/bNWUzG/1/
                    if unix_socket_match:
                        params['host'] = unix_socket_match.group('path')
                        params['port'] = int(unix_socket_match.group('port'))
                    # end if
                    psycopg2_conn = psycopg2_connect(**params)
                # end if

                placeholder_values = [sql_escaping.Literal(value) for value in placeholder_values]
                formatting_dict = {f'default_placeholder_{i}': val for i, val in enumerate(placeholder_values)}
                sql = sql_escaping.SQL(sql)
                sql = sql.format(**formatting_dict).as_string(context=psycopg2_conn)
            # end if
        # noinspection PyRedundantParentheses
        return (sql, *[])
    # end def

    @classmethod
    def match_type(
        cls,
        type_hint: TYPEHINT_TYPE,
        *,
        is_automatic_field: Optional[bool] = None,
        key: Optional[str] = None,
        is_outer_call: bool = True
    ) -> Tuple[bool, str]:
        """
        Processes a type hint to produce a CREATE TABLE sql segment of the type of that type hint and if it's optional..

            >>> class Example(FastORM):
            ...   foo: Optional[int]
            ...

            >>> type_hints = typing.get_type_hints(Example)
            >>> Example.match_type(type_hints['foo'], is_automatic_field=False)
            (True, 'BIGINT')

        """
        is_union_type = check_is_union_type(type_hint)
        if hasattr(type_hint, '__origin__') or is_union_type:
            if check_is_annotated_type(type_hint):
                # https://stackoverflow.com/q/68275615/3423324#what-is-the-right-way-to-check-if-a-type-hint-is-annotated
                actual_type = type_hint.__origin__  # str or wherever was the first parameter.
                metadata = type_hint.__metadata__
                if not isinstance(metadata, (tuple, list)):
                    metadata = (metadata,)
                # end if
                if AutoincrementType in metadata:
                    is_automatic_field = True
                # end if
                is_optional, sql_type = cls.match_type(
                    actual_type, is_automatic_field=is_automatic_field, is_outer_call=False
                )
                return is_optional, sql_type
            # end if

            origin = type_hint.__origin__ if hasattr(type_hint, '__origin__') else type(type_hint)
            is_union_type = check_is_union_type(origin)
            if is_union_type or origin in (typing.Optional, typing.Union):  # Optional is an special union, too
                union_params = type_hint.__args__[:]  # this was __union_params__ in python3.5, but __args__ in 3.6+
                if not isinstance(union_params, (list, tuple)):
                    raise TypeError(
                        f'Union type for key {key} has unparsable params.', union_params,
                    )
                # end if
                if NoneType in union_params:
                    is_optional = True
                    union_params = [param for param in union_params if not issubclass(param, NoneType)]
                else:
                    is_optional = False
                # end if
                if len(union_params) == 0:
                    raise TypeError(
                        f'Union with no (non-None) type(s) at key {key}.', type_hint.__args__,
                    )
                # end if
                first_union_type = union_params[0]
                if not all(first_union_type == x for x in union_params[1:]):
                    raise TypeError(
                        f'Union with more than one type at key {key}.', union_params,
                    )
                # end if
                additional_is_optional, sql_type = cls.match_type(
                    first_union_type, is_automatic_field=is_automatic_field, is_outer_call=False
                )
                if additional_is_optional:
                    is_optional = True
                # end if
            elif isinstance(origin, typing.List) or issubclass(origin, list):
                list_params = type_hint.__args__
                if len(list_params) != 1:  # list has one type
                    raise TypeError(
                        'List with more than one type parameter.', type_hint, list_params
                    )
                # end if
                the_type = list_params[0]
                try:
                    # we will now recursively go into that list.
                    # if it is like `list[list[list[int]]] it will succeed as INT,
                    # and for those 3 lists the [] will be added 3 times, resulting in INT[][][]
                    # If any of those inner lists aren't a compatible type (TypeError),
                    # e.g. list[list[Union[str, int]]], we have to use a json dict instead.
                    _, sql_type = cls.match_type(
                        the_type, is_automatic_field=is_automatic_field, is_outer_call=False
                    )
                    sql_type = "".join((sql_type, "[]"))  # append '[]' to the sql_type
                    return False, sql_type  # the list itself can't be optional, that has to be done by an outer Optional[].
                except TypeError as e:
                    if not is_outer_call:
                        # make sure we don't end up with JSONB[] for list[list[Union[str, int]]],
                        # only the outer one should migrate to json.
                        raise e
                    # end if
                    logger.debug('Could not parse as a single type list (e.g. INT[][]), now will be a json field.', exc_info=True)
                    return False, cls._COLUMN_TYPES[dict]
                # end try
            elif isinstance(origin, typing.Tuple) or issubclass(origin, builtins.tuple):
                tuple_params = type_hint.__args__
                if len(tuple_params) == 0:  # list has one type
                    raise TypeError(
                        'Tuple has no parameters.', type_hint, tuple_params
                    )
                # end if

                # check if all types of the tuple are the same, so we can use a list
                first_type = tuple_params[0]
                if all(first_type == x for x in tuple_params[1:]):
                    # we hope this will give us something like  INT, TEXT, FLOAT, etc.
                    _, sql_type = cls.match_type(
                        first_type, is_automatic_field=is_automatic_field, is_outer_call=False
                    )
                    sql_type = "".join((sql_type, "[]"))  # append '[]' to the sql_type
                    return False, sql_type  # the tuple itself can't be optional, that has to be done by an outer Optional[].
                # end if

                # so the types are all over the place, so we will have to fallback to json.
                sql_type = cls._COLUMN_TYPES[dict]
                return False, sql_type  # the list itself can't be optional, that has to be done by an outer Optional[].
            else:
                raise ValueError('Enclosed by an unknown type', origin, f'key={key!r}')
            # end case
        elif isinstance(type_hint, ModelField):
            is_optional = type_hint.allow_none
            # is_optional = type_hint.allow_none and (type_hint.shape != SHAPE_SINGLETON or not type_hint.sub_fields)
            try:
                subtype_is_optional, sql_type = cls.match_type(
                    type_hint=type_hint.type_, is_automatic_field=is_automatic_field, key=key, is_outer_call=False,
                )
            except TypeError as e:
                if not is_outer_call:
                    # make sure we don't end up with JSONB[] for list[list[Union[str, int]]],
                    # only the outer one should migrate to json.
                    raise e
                # end if
                logger.debug(
                    'Could not parse as a single type list (e.g. INT[][]), now will be a json field.', exc_info=True
                )
                return False, cls._COLUMN_TYPES[dict]
            # end try

            # pydantic makes t6_2: str = None  to be  t6_2: Optional[str] = None, couldn't find a way to detect only the first variant.
            # if is_optional and type_hint.field_info.default is None:
            #     # e.g. t6_2: str = None
            #     assert type_hint.field_info.default is None
            #     assert not isinstance(type_hint.field_info.default, UndefinedType)
            #     raise ValueError("You can't have an non-optional type default to None")
            # # end if

            if type_hint.outer_type_ == type_hint.type_:
                # e.g. for str,  typehint.outer_type_ is str
                #      for str,  typehint.type_       is str
                # but also for Optional[str] it is  both  str!
                return is_optional, sql_type
            if check_is_generic_alias(type_hint.outer_type_) and hasattr(type_hint.outer_type_, '__origin__'):
                # e.g. for list[int],  typehint.outer_type_ is List[int], thus having a .__origin__ == list
                #      for list[int],  typehint.type_       is int
                # isinstance(list[int], GenericAlias) == True

                wrapper_class = type_hint.outer_type_.__origin__  # e.g. list if we had List[int].
                if issubclass(wrapper_class, list):
                    sql_type = "".join([sql_type, "[]"])
                # end if
            elif check_is_annotated_type(type_hint.outer_type_):
                return cls.match_type(
                    type_hint.outer_type_, is_automatic_field=is_automatic_field, is_outer_call=False
                )
            # end if
        else:
            is_optional = False
            sql_type = cls._match_type(type_hint, automatic=is_automatic_field)  # fails anyway if not in the list above
        # end case
        return is_optional, sql_type
    # end def

    @classmethod
    def _param_is_list_of_multiple_values(cls, key: str, value: Any, typehint: Any):
        """
        If a value is multiple times of what was defined.
        :param key:
        :param value:
        :param typehint:
        :return: True if the `value` is a list of tuple of arguments satisfying the `typehint`.
        """
        if not isinstance(value, (list, tuple)):
            # we don't have a list -> can't be multiple values
            # this is a cheap check preventing most of the values.
            return False
        # end if

        original_type_fits = False
        listable_type_fits = False
        try:
            check_type(argname=key, value=value, expected_type=typehint)
            original_type_fits = True  # the original was already compatible
        except TypeError:
            pass
        # end if
        try:
            check_type(argname=key, value=value, expected_type=Union[Tuple[typehint], List[typehint]])
            listable_type_fits = True  # the original was already compatible
        except TypeError:
            pass
        # end if

        logger.debug(f'type fitting analysis: original: {original_type_fits}, tuple/list: {listable_type_fits}')
        if not listable_type_fits:
            # easy one,
            # so our list/tuple check doesn't match,
            # so it isn't compatible.
            return False
        # end if
        if not original_type_fits:
            # tuple/list one fits
            # but the original one does not.
            # pretty confident we have a list/tuple of a normal attribute and can do a `WHERE {key} IN ({",".join(value)})`
            return True
        # end if

        # else -> listable_type_fits == True, original_type_fits == True
        # ooof, so that one would have already be accepted by the original type...
        # That's a tough one...
        # we err to the side of caution
        logger.warn(f'Not quite sure if it fits or not, erring to the side of caution and assuming single parameter.')
        return False
    # end def

    _CLASS_SERIALIZERS = {
        # # class: callable(data) -> json
        # # E.g.:
        # TgBotApiObject: lambda obj: return obj._raw if hasattr(obj, '_raw') and obj._raw else obj.to_array()
    }

    @classmethod
    async def create_connection(cls, database_url) -> Connection:
        # https://magicstack.github.io/asyncpg/current/usage.html#example-automatic-json-conversion
        conn = await asyncpg.connect(database_url)
        return await cls._set_up_connection(conn=conn)
    # end def

    @classmethod
    async def create_connection_pool(cls, database_url) -> Pool:
        # https://magicstack.github.io/asyncpg/current/usage.html#example-automatic-json-conversion
        return await asyncpg.create_pool(database_url, setup=cls._set_up_connection)
    # end def


    @classmethod
    async def _set_up_connection(cls, conn: Connection):
        """
        Sets up a connection to properly do datetime and json decoding.

        Prepares writing datetimes (as ISO format) and class instances as json if they have a `.to_dict()`, `.to_array()` function.
        An easy way to add your is by having a `.to_json()` function like above or
        appending your class to `_CLASS_SERIALIZERS` like so:
        ```py
        # anywhere in your code, to be run once
        FastORM._CLASS_SERIALIZERS[SomeClass] = lambda obj: obj.do_something()
        ```
        :param conn:
        :return:
        """
        import json

        def decoder_with_empty(text):
            if text.strip() == '':
                return None
            # end if
            return json.loads(text)
        # end def

        def json_dumps(obj):
            logger.debug(f'Encoding to JSON: {obj!r}')

            def my_converter(o):
                if isinstance(o, datetime.datetime):
                    return o.isoformat()
                # end def
                if hasattr(o, 'to_array'):
                    return o.to_array()  # TgBotApiObject from pytgbot
                # end def
                if hasattr(o, 'to_dict'):
                    return o.to_dict()
                # end def

                # check _CLASS_SERIALIZERS,
                # a easy way to add your own by writing FastORM._CLASS_SERIALIZERS[Class] = lambda obj: obj
                for type_to_check, callable_function in cls._CLASS_SERIALIZERS.items():
                    if isinstance(o, type_to_check):
                        return callable_function(o)
                    # end if
                # end for
            # end def
            return json.dumps(obj, default=my_converter)
        # end def

        for sql_type in ('json', 'jsonb'):
            await conn.set_type_codec(
                sql_type,
                encoder=json_dumps,
                decoder=decoder_with_empty,
                schema='pg_catalog'
            )
        # end for
        return conn
    # end def


    @classmethod
    async def get_connection(cls, database_url):
        """
        Deprecated, use `FastORM.create_connection(…)` instead.
        """
        logger.warning(
            'The use of FastORM.get_connection(…) is deprecated, use FastORM.create_connection(…)` instead.',
            exc_info=True,
        )
        return await cls.create_connection(database_url=database_url)
    # end def
# end class


class AutoincrementType(object):
    """
    Below there will be a singleton called Autoincrement.
    It can be used in the following two ways, they are the same:

        >>. Autoincrement = AutoincrementType()
        >>. class Foo(FastORM):
        ...   id_a: int = Field(default_factory=Autoincrement, other_field_parameters=...)
        ...   id_b: int = Autoincrement(other_field_parameters=...)

    In other words, the following two are doing the same:

        >>. example_1 = Autoincrement(default=None)
        >>. example_2 = Field(default_factory=Autoincrement)

        >>. from pydantic.fields import FieldInfo
        >>. isinstance(example_1, FieldInfo)
        True
        >>. isinstance(example_2, FieldInfo)
        True

        >>. Autoincrement().__dir__() == Field(default_factory=Autoincrement).__dir__()
        True

    """

    def __init__(self):
        self.__name__ = "fastorm.Autoincrement"  # some internals of Field want to know that.
    # end def

    @typing.overload
    def __call__(
        self,
        default: Any = Undefined,
        *,
        default_factory: Optional[NoArgAnyCallable] = None,
        alias: str = None,
        title: str = None,
        description: str = None,
        const: bool = None,
        gt: float = None,
        ge: float = None,
        lt: float = None,
        le: float = None,
        multiple_of: float = None,
        min_items: int = None,
        max_items: int = None,
        min_length: int = None,
        max_length: int = None,
        allow_mutation: bool = True,
        regex: str = None,
        **extra: Any,
    ) -> Any:
        pass
    # end def

    def __call__(self, *args, **kwargs) -> None:
        if not args and not kwargs:
            # For the use in `Field(default_factory=Autoincrement)`, we will be called without parameters.
            # so we return the default value this object should then have, that is `None`.
            # It would be cooler, but we can't return Autoincrement as that would be incompatible with the field's type.
            return None
        # end if

        # so it's not the `Field(default_factory=Autoincrement)` calling us with zero parameters
        assert 'default_factory' not in kwargs
        kwargs['default_factory'] = Autoincrement
        return Field(*args, **kwargs)
    # end def
# end class


Autoincrement = AutoincrementType()



import os
import unittest
from datetime import datetime
from typing import Optional, Any, Type
from pydantic import dataclasses
from pydantic.fields import Undefined, Field

from fastorm import FastORM, Autoincrement
from tests.tools_for_the_tests_of_fastorm import extract_sql_from_docstring

POSTGRES_DSN_URL = os.getenv('POSTGRES_DSN_URL', 'postgres://')  # default is to try localhost
assert POSTGRES_DSN_URL is not None


try:
    from unittest import IsolatedAsyncioTestCase
    has_easy_unittest = True
    unittest_cls = IsolatedAsyncioTestCase
except ImportError:
    has_easy_unittest = False
    unittest_cls = unittest.TestCase
# end if


@dataclasses.dataclass
class ExpectedResult(object):
    is_optional: bool
    sql_type: str
    default: Any
# end class

ExpectedResult: Type[Any]


DATETIME_NOW = datetime(4458, 12, 24, 0, 6, 9)

class SystemUnderTest(FastORM):
    """
        CREATE TABLE "cool_table_yo" (
          "t0_id" BIGSERIAL NOT NULL PRIMARY KEY,
          "t1_1" TEXT NOT NULL DEFAULT 'test',
          "t1_2" TEXT DEFAULT NULL,
          "t1_3" BIGINT NOT NULL DEFAULT 4458,
          "t1_4" TEXT NOT NULL DEFAULT 'this test will proof if "something" ain''t escaped properly. ^^''',
          "t2_1" TEXT NOT NULL DEFAULT 'test ööö',
          "t3_1" TIMESTAMP NOT NULL DEFAULT '4458-12-24T00:06:09'::timestamp
        );
    """
    _table_name = 'cool_table_yo'
    _primary_keys = ['t0_id']
    _automatic_fields = ['t0_id']
    _ignored_fields = []

    t0_id: int = Field(default_factory=Autoincrement)
    __result__t0_id = ExpectedResult(is_optional=False, sql_type="BIGINT", default=Undefined)

    #
    # defaults
    #

    t1_1: str = "test"
    __result__t1_1 = ExpectedResult(is_optional=False, sql_type="TEXT", default="test")

    t1_2: Optional[str] = None
    __result__t1_2 = ExpectedResult(is_optional=True, sql_type="TEXT", default=None)

    t1_3: int = 4458
    __result__t1_3 = ExpectedResult(is_optional=False, sql_type="BIGINT", default=4458)

    t1_4: str = "this test will proof if \"something\" ain't escaped properly. ^^'"
    __result__t1_4 = ExpectedResult(is_optional=False, sql_type="TEXT", default="this test will proof if \"something\" ain't escaped properly. ^^'")

    #
    # Non-Ascii str
    #


    t2_1: str = 'test ööö'
    __result__t2_1 = ExpectedResult(is_optional=False, sql_type="TEXT", default='test ööö')

    #
    # other types
    #

    t3_1: datetime = DATETIME_NOW
    __result__t3_1 = ExpectedResult(is_optional=True, sql_type="DATETIME", default=DATETIME_NOW)
# end class


class CreateTableOnlineTestCase(unittest_cls):
    def test_sql_text_connection_missing(self):
        with self.assertRaises(ValueError) as e:
            SystemUnderTest.build_sql_create()
        # end with
        self.assertEqual(str(e.exception), 'For using complex default values (everything other than None, bool, int, and pure ascii strings) a psycopg2 connection or cursor needs to be provided as the psycopg2_conn parameter.')
    # end def

    def test_sql_text_connection_typeerror(self):
        with self.assertRaises(TypeError) as e:
            SystemUnderTest.build_sql_create(psycopg2_conn=127834)
        # end with
        self.assertEqual(str(e.exception), 'For using complex default values (everything other than None, bool, int, and pure ascii strings) a psycopg2 connection or cursor needs to be provided as the psycopg2_conn parameter.')
    # end def

    def test_sql_text_connection_valid_psycop2(self):
        expected_sql = extract_sql_from_docstring(SystemUnderTest)
        import psycopg2
        connection = psycopg2.connect(POSTGRES_DSN_URL)
        actual_sql, *actual_params = SystemUnderTest.build_sql_create(psycopg2_conn=connection)
        self.assertEqual(expected_sql, actual_sql)
        self.assertListEqual([], actual_params)
    # end def

    async def test_sql_text_connection_valid_asyncpg(self):
        expected_sql = extract_sql_from_docstring(SystemUnderTest)

        import asyncpg
        connection = await asyncpg.connect(POSTGRES_DSN_URL)
        actual_sql, *actual_params = SystemUnderTest.build_sql_create(psycopg2_conn=connection)
        self.assertEqual(expected_sql, actual_sql)
        self.assertListEqual([], actual_params)
    # end def
# end class


if __name__ == '__main__':
    unittest.main()
# end if



import unittest
from typing import Optional, Union, Any, Type, List, Tuple, Dict

from fastorm import FastORM
from tests.tools_for_the_tests_of_fastorm import extract_sql_from_docstring


class Table1(FastORM):
    _table_name = 'table1'
    _automatic_fields = ['id']
    _primary_keys = ['id']

    id: Union[int, None]
# end class


class Table2(FastORM):
    _table_name = 'table2'
    _automatic_fields = ['id']
    _primary_keys = ['id']

    id: Union[int, None]
# end class


class Table1HavingTable2VersionSingleReferencesOptional(FastORM):
    """
        CREATE TABLE "table1_having_table2" (
          "table1__id" BIGINT,
          "table2__id" BIGINT,
          PRIMARY KEY ("table1__id", "table2__id")
        );
        CREATE INDEX "idx_table1_having_table2___table1__id" ON "table1_having_table2" ("table1__id");
        CREATE INDEX "idx_table1_having_table2___table2__id" ON "table1_having_table2" ("table2__id");
        ALTER TABLE "table1_having_table2" ADD CONSTRAINT "fk_table1_having_table2___table1__id" FOREIGN KEY ("table1__id") REFERENCES "table1" ("id") ON DELETE CASCADE;
        ALTER TABLE "table1_having_table2" ADD CONSTRAINT "fk_table1_having_table2___table2__id" FOREIGN KEY ("table2__id") REFERENCES "table2" ("id") ON DELETE CASCADE;
    """
    _table_name = 'table1_having_table2'
    _primary_keys = ['table1', 'table2']
    _automatic_fields = ['table1', 'table2']

    table1: Union[Table1, int]
    table2: Union[Table2, int]
# end class


class Table1HavingTable2VersionSingleReferencesMandatory(FastORM):
    """
        CREATE TABLE "table1_having_table2" (
          "table1__id" BIGINT,
          "table2__id" BIGINT,
          PRIMARY KEY ("table1__id", "table2__id")
        );
        CREATE INDEX "idx_table1_having_table2___table1__id" ON "table1_having_table2" ("table1__id");
        CREATE INDEX "idx_table1_having_table2___table2__id" ON "table1_having_table2" ("table2__id");
        ALTER TABLE "table1_having_table2" ADD CONSTRAINT "fk_table1_having_table2___table1__id" FOREIGN KEY ("table1__id") REFERENCES "table1" ("id") ON DELETE CASCADE;
        ALTER TABLE "table1_having_table2" ADD CONSTRAINT "fk_table1_having_table2___table2__id" FOREIGN KEY ("table2__id") REFERENCES "table2" ("id") ON DELETE CASCADE;
    """
    _table_name = 'table1_having_table2'
    _primary_keys = ['table1', 'table2']
    _automatic_fields = ['table1', 'table2']

    table1: Table1
    table2: Table2
# end class


class DoublePrimaryKeyTable(FastORM):
    """
        CREATE TABLE "table1_having_table2" (
          "table1__id" BIGINT NOT NULL,
          "table2__id" BIGINT,
          PRIMARY KEY ("id_part1", "id_part2")
        );
    """
    _table_name = 'double_primary_key'
    _primary_keys = ['id_part1', 'id_part2']
    _automatic_fields = []

    id_part1: int
    id_part2: Optional[int]
# end class


class ReferencingDoublePrimaryKeyTableVersionMultiReferencesMandatory(FastORM):
    """
        CREATE TABLE "double_primary_key" (
          "double_trouble__id_part1" BIGINT NOT NULL,
          "double_trouble__id_part2" BIGINT,
          PRIMARY KEY ("double_trouble__id_part1", "double_trouble__id_part2")
        );
        CREATE INDEX "idx_double_primary_key___double_trouble__id_part1" ON "double_primary_key" ("double_trouble__id_part1");
        CREATE INDEX "idx_double_primary_key___double_trouble__id_part2" ON "double_primary_key" ("double_trouble__id_part2");
        ALTER TABLE "double_primary_key" ADD CONSTRAINT "fk_double_primary_key___double_trouble__id_part1" FOREIGN KEY ("double_trouble__id_part1") REFERENCES "double_primary_key" ("id_part1") ON DELETE CASCADE;
        ALTER TABLE "double_primary_key" ADD CONSTRAINT "fk_double_primary_key___double_trouble__id_part2" FOREIGN KEY ("double_trouble__id_part2") REFERENCES "double_primary_key" ("id_part2") ON DELETE CASCADE;
    """
    _table_name = 'double_primary_key'
    _primary_keys = ['double_trouble']
    _automatic_fields = []

    double_trouble: DoublePrimaryKeyTable
# end class


class ReferencingDoublePrimaryKeyTableVersionMultiReferencesOptional(FastORM):
    """
        CREATE TABLE "double_primary_key" (
          "double_trouble__id_part1" BIGINT NOT NULL,
          "double_trouble__id_part2" BIGINT,
          PRIMARY KEY ("double_trouble__id_part1", "double_trouble__id_part2")
        );
        CREATE INDEX "idx_double_primary_key___double_trouble__id_part1" ON "double_primary_key" ("double_trouble__id_part1");
        CREATE INDEX "idx_double_primary_key___double_trouble__id_part2" ON "double_primary_key" ("double_trouble__id_part2");
        ALTER TABLE "double_primary_key" ADD CONSTRAINT "fk_double_primary_key___double_trouble__id_part1" FOREIGN KEY ("double_trouble__id_part1") REFERENCES "double_primary_key" ("id_part1") ON DELETE CASCADE;
        ALTER TABLE "double_primary_key" ADD CONSTRAINT "fk_double_primary_key___double_trouble__id_part2" FOREIGN KEY ("double_trouble__id_part2") REFERENCES "double_primary_key" ("id_part2") ON DELETE CASCADE;
    """
    _table_name = 'double_primary_key'
    _primary_keys = ['double_trouble']
    _automatic_fields = []

    double_trouble: Union[DoublePrimaryKeyTable, Tuple[int, int]]
# end class


# ----------------------------------------------------


class CreateTableTestCase(unittest.TestCase):
    def test_working_table_single_references_mandatory(self):
        cls = Table1HavingTable2VersionSingleReferencesMandatory
        expected_sql = extract_sql_from_docstring(cls)
        actual_sql, *actual_params = cls.build_sql_create()
        self.assertEqual(expected_sql, actual_sql)
    # end def

    def test_working_table_single_references_optional(self):
        cls = Table1HavingTable2VersionSingleReferencesOptional
        expected_sql = extract_sql_from_docstring(cls)
        actual_sql, *actual_params = cls.build_sql_create()
        self.assertEqual(expected_sql, actual_sql)
    # end def

    def test_working_table_multi_references_mandatory(self):
        cls = ReferencingDoublePrimaryKeyTableVersionMultiReferencesMandatory
        expected_sql = extract_sql_from_docstring(cls)
        actual_sql, *actual_params = cls.build_sql_create()
        self.assertEqual(expected_sql, actual_sql)
    # end def

    def test_working_table_multi_references_optional(self):
        cls = ReferencingDoublePrimaryKeyTableVersionMultiReferencesOptional
        expected_sql = extract_sql_from_docstring(cls)
        actual_sql, *actual_params = cls.build_sql_create()
        self.assertEqual(expected_sql, actual_sql)
    # end def
# end class


if __name__ == '__main__':
    unittest.main()
# end if



import unittest
from datetime import datetime
from typing import get_type_hints
from typing import Optional, Union, Any, Type, List, Tuple, Dict
from pydantic import dataclasses, BaseModel
from pydantic.fields import ModelField, Undefined, Field

from fastorm import FastORM
from fastorm.compat import get_type_hints_with_annotations
from tests.tools_for_the_tests_of_fastorm import extract_sql_from_docstring


class OtherTable(FastORM):
    """
        CREATE TABLE "other_table" (
          "id_part_1" TEXT NOT NULL,
          "id_part_2" BIGINT NOT NULL,
          PRIMARY KEY ("id_part_1", "id_part_2")
        );
    """
    _table_name = 'cool_table_name_yoooo'
    _primary_keys = ['id_part_1', 'id_part_2']
    _ignored_fields = []

    id_part_1: int
    id_part_2: str
    foo: float
    bar: str
# end class


class SelectRowTestCase(unittest.TestCase):
    def test_simple_select_single_parameter_no_pk(self):
        expected_foo = 123.2
        # noinspection SqlResolve,SqlNoDataSourceInspection
        expected_sql = """
        SELECT "id_part_1","id_part_2","foo","bar" FROM "cool_table_name_yoooo" WHERE "foo" = $1
        """.strip()
        sql, *where_values = OtherTable.build_sql_select(foo=expected_foo)
        self.assertEqual(expected_sql, sql)
        self.assertEqual([expected_foo], where_values)
    # end def

    def test_simple_select_multiple_parameters_no_pk(self):
        expected_foo = 123.2
        expected_bar = "Läääl"
        # noinspection SqlResolve,SqlNoDataSourceInspection
        expected_sql = """
        SELECT "id_part_1","id_part_2","foo","bar" FROM "cool_table_name_yoooo" WHERE "foo" = $1 AND "bar" = $2
        """.strip()
        sql, *where_values = OtherTable.build_sql_select(foo=expected_foo, bar=expected_bar)
        self.assertEqual(expected_sql, sql)
        self.assertEqual([expected_foo, expected_bar], where_values)
    # end def

    def test_simple_select_no_parameters_no_pk(self):
        # noinspection SqlResolve,SqlNoDataSourceInspection
        expected_sql = """
        SELECT "id_part_1","id_part_2","foo","bar" FROM "cool_table_name_yoooo"
        """.strip()
        sql, *where_values = OtherTable.build_sql_select()
        self.assertEqual(expected_sql, sql)
        self.assertEqual([], where_values)
    # end def
# end class


if __name__ == '__main__':
    unittest.main()
# end if



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'luckydonald'

import sys
from typing import Any, Union, Dict
from pydantic.fields import ModelField


TYPEHINT_TYPE = Union[type, ModelField]

IS_PYTHON_3_7 = sys.version_info[:3] >= (3, 7, 0)
IS_PYTHON_3_10 = sys.version_info[:3] >= (3, 10, 0)


# noinspection PyUnusedLocal
def check_is_union_type(variable: Any) -> bool:
    # as there's no UnionType, we can't have an instance of it.
    return False
# end def


def check_is_generic_alias(variable: Any) -> bool:
    # as there's no GenericAlias, we can't have an instance of it.
    return False
# end def


def check_is_annotated_type(variable: Any) -> bool:
    # as there's no Annotated, we can't have an instance of it.
    return False
# end def


try:
    from types import UnionType

    def check_is_union_type(variable: Any) -> bool:
        return isinstance(variable, UnionType)
    # end def

    TYPEHINT_TYPE = Union[TYPEHINT_TYPE, UnionType]
except ImportError:
    pass
# end try


try:
    from types import GenericAlias
    from typing import List

    GenericAliasOld = type(List[int])

    def check_is_generic_alias(variable: Any) -> bool:
        return (
            isinstance(variable, GenericAlias) or  # type(list[int])
            isinstance(variable, GenericAliasOld)  # type(List[int])
        )
    # end def

    TYPEHINT_TYPE = Union[TYPEHINT_TYPE, GenericAlias, GenericAliasOld]
except ImportError:
    pass
# end try

Annotated = None

try:
    try:
        from typing import Annotated  # 3.10
    except ImportError:
        from typing_extensions import Annotated  # 3.7
    # end try

    AnnotatedType = type(Annotated)


    def check_is_annotated_type(variable: Any) -> bool:
        # as there's no Annotated, we can't have an instance of it.
        return isinstance(variable, AnnotatedType)
    # end def
except ImportError:
    pass
# end try

try:
    from types import NoneType
except:
    NoneType = type(None)
# end try

if IS_PYTHON_3_10:
    from typing import get_type_hints

    def get_type_hints_with_annotations(cls) -> Dict[str, any]:
        return get_type_hints(cls, include_extras=True)
    # end def
elif not IS_PYTHON_3_7:
    from typing import get_type_hints

    def get_type_hints_with_annotations(cls) -> Dict[str, any]:
         return get_type_hints(cls, include_extras=True)
    # end def
else:  # pre 3.7
    from typing import get_type_hints

    def get_type_hints_with_annotations(cls) -> Dict[str, any]:
         return get_type_hints(cls)
    # end def
# end def




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'luckydonald'

import asyncio
from datetime import datetime, timedelta
from typing import Optional, Union
from enum import Enum

from asyncpg import Connection
from luckydonaldUtils.typing import JSONType

from fastorm import FastORM


class State(str, Enum):
    RUNNING = 'running'
    ABORTED = 'stopped'
    COMPLETED = 'done'
# end class


class User(FastORM):
    _ignored_fields = []
    _primary_keys = ['id']
    _automatic_fields = ['id']
    _table_name = 'user'

    id: Optional[int]  # Optional because automatically filled (_automatic_fields)
    name: str
# end class


class Auction(FastORM):
    _ignored_fields = []
    _primary_keys = ['id']
    _automatic_fields = ['id']
    _table_name = 'auction'

    id: Optional[int]  # Optional because automatically filled (_automatic_fields)
    owner: Union[int, User]  # can be provided by int or the native object
    previous_owner: Optional[User]  # Optional because nullable
    state: State  # Enum support :D
    title: str
    subtitle: Optional[str]  # nullable
    description: str
    start_date: datetime  # datetime support
    end_date: datetime
    metadata: JSONType
    deleted: bool
    chat_id: int
# end class


async def run():
    conn = await FastORM.create_connection('postgresql://user:password@postgres_host/database')

    await migrate_database(conn)  # a function (below) which prepares the database for use, creates tables and manages schema changes.

    owner_user = User(id=1234, name="The new auction owner.")
    await owner_user.insert(conn=conn, ignore_setting_automatic_fields=True)

    previous_owner = User(id=None, name="The previous auction owner.")
    await previous_owner.insert(conn=conn, ignore_setting_automatic_fields=False)

    print(f'Inserted owner user as id={owner_user.id}.')
    print(f'Inserted previous owner user as id={previous_owner.id}.')

    auction = Auction(
        id=None,  # gonna be automatic if `insert(…, ignore_setting_automatic_fields=False)` (default).
        owner=1234,  # by numeric id, aka. the actual field value
        previous_owner=previous_owner,  # use the id field (`User._primary_keys`) to determine the actual integer value.
        state=State.RUNNING,  # enum will be a string in the database
        title="I sell my soul", subtitle="Slightly used",
        description="You only get a share though.",
        start_date=datetime.now(), end_date=datetime.now() + timedelta(days=5),  # datetimes just works
        metadata={"just": ["json", "stuff", 111, datetime.now()]},  # will be native JSONB. You can have datetimes and your own classes in there as well, see `FastORM._set_up_connection`.
        deleted=False,
        chat_id=9223372036854775807,  # note, this database field must be BIGINT for such large numbers
    )

    await auction.insert(conn=conn, ignore_setting_automatic_fields=True, on_conflict_upsert_field_list=None,  write_back_automatic_fields=True)
    print(f'Inserted auction as id={auction.id}.')
    auction_id = auction.id

    # now let's retrieve some data from the database!

    db_auctions_by_user = await Auction.select(conn=conn, owner=1234)  # you can provide everything after the `conn=conn` to identify your entries. Returns a list, which can be empty.
    db_auction = await Auction.get(conn=conn, id=auction_id)  # same as select, but only allows a single result. None if not found. Raises an error if it's more than one match.
    print(db_auction.title)

    # let's change something.
    db_auction.state = State.COMPLETED
    await db_auction.update(conn=conn)

    # run custom queries

    # e.g. joining the users to that result and loading them in a single query
    # basically we use a prepared statement for variables.
    fetch_params = (
        f'SELECT '
        f' {Auction.get_select_fields(namespace="auction")}'  # first all fields from the auction
        f', {User.get_select_fields(namespace="owner")}'  # then all fields from the owner
        f' FROM'
        f'  {Auction.get_table()} AS "auction",'  # notice this is the same as the `namespace=…` above
        f'  {User.get_table()} AS "owner",'  # notice this is the same as the `namespace=…` above
        # now the join condition for those table
        f'WHERE "owner"."id" = "auction"."owner"'  # join user table with auction for owner
        # and more general condition we are interested right now
        f' AND "auction"."end_date" > CURRENT_TIMESTAMP'  # still running
        f' AND ("auction"."state" = $1 OR "auction"."id" = $2)'  # I don't know, but two parameters so I can demostrate that.
        f' ORDER BY "auction"."end_date"'  # ascending, so we have the soonest to end first
        ,
        State.COMPLETED,
        1234,
    )  # so to make it clear, syntax is ["SQL", $1, $2, …], list or tuple
    # now run that query
    fetch_results = await conn.fetch(*fetch_params)
    # now we can deconstruct that into rows
    auction_fields_len = Auction.get_select_fields_len()
    user_fields_len = User.get_select_fields_len()
    for row in fetch_results:
        # split the results into the proper lists for both elements
        auction_fields = row[0:auction_fields_len]
        user_fields = row[auction_fields_len:auction_fields_len + user_fields_len]

        # generate objects from it
        database_auction = Auction.from_row(auction_fields)
        database_user = User.from_row(user_fields)
        print(f'Auction {database_auction.title} from user {database_user.name}')
    # end for
# end def


async def migrate_database(conn: Connection):
    """
    Database migration, to set up all the tables we need. With versions so you can simply add a new version if you modify your models.
    Usually a separate migration file, just in the same one here for the sake of the tutorial being a single file.
    """
    await conn.execute(
        """
        CREATE TABLE IF NOT EXISTS "version" (
          "version" SERIAL PRIMARY KEY,
          "date" TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
          "message" TEXT NOT NULL,
          "details" TEXT NOT NULL,
          "updated_from_version" INTEGER
        )
        """
    )
    # Select the version from the table and handle the table being just created as version 0.
    now = datetime.now()
    row = await conn.fetchrow('SELECT "version", "date", "message", "details" FROM "version" ORDER BY "version" DESC')
    if row is None:
        version = None
        print(f'database no version yet, assuming version 0.')
    else:
        version, date, message, details = row
        print(f'database has version {version} from {date}: {message} ({details})')
    # end if
    updated_from_version = version

    if version < 1:
        version = 1
        message = "Create a first update, with a user table"
        details = "Just an user for now and the auction table later for the purpose of a tutorial"
        await conn.execute(
            """
            CREATE TABLE "user" (
                "id" SERIAL PRIMARY KEY,
                "name" TEXT NOT NULL
            )
            """
        )
        await conn.execute(
            '''INSERT INTO version(version, date, message, details, updated_from_version) VALUES($1, $2, $3, $4, $5)''',
            version, now, message, details, updated_from_version
        )
    # end if

    if version < 2:
        version = 2
        message = "Create a second update, with the auction table"
        details = None
        await conn.execute(
            """
            CREATE TABLE IF NOT EXISTS "auction" (

              "id" SERIAL PRIMARY KEY,  -- id: Optional[int]  # Optional because automaticly filled (_automatic_fields)
              "category" TEXT NOT NULL,
              "owner" INTEGER NOT NULL,  -- owner: Union[int, User]  # can be provided by int or the native object
              "previous_owner" INTEGER,  -- previous_owner: Optional[User]  # Optional because nullable
              "state" TEXT NOT NULL, -- state: State  # Enum support :D
              "title" TEXT NOT NULL, -- title: str
              "subtitle" TEXT, -- subtitle: Optional[str]  # nullable
              "description" TEXT NOT NULL, -- description: str
              "start_date" TIMESTAMP NOT NULL, -- start_date: datetime  # datetime support
              "end_date" TIMESTAMP NOT NULL, -- end_date: datetime
              "metadata" JSONB NOT NULL, -- metadata: JSONType
              "deleted" BOOLEAN NOT NULL, -- deleted: bool
              "chat_id" BIGINT NOT NULL  -- chat_id: int
            );
            """
        )
        await conn.execute(
            '''INSERT INTO version(version, date, message, details, updated_from_version) VALUES($1, $2, $3, $4, $5)''',
            version, now, message, details, updated_from_version
        )
    # end if
    print(f'Updated database from {updated_from_version} to {version}')
# end def


# Spin up a quick and simple event loop
# and run until completed
loop = asyncio.get_event_loop()
try:
    loop.run_until_complete(run())
finally:
    loop.close()
# end try



import unittest

from test_create_table import SystemUnderTest, OtherTable, TheReferenceHasBeenDoubled


class TableMetadataTestCase(unittest.TestCase):
    def test_table_quoted_name(self):
        tables = {
            SystemUnderTest: '"cool_table_yo"',
            OtherTable: '"other_table"',
            TheReferenceHasBeenDoubled: '"double_reference"',
        }
        for table, expected_name in tables.items():
            with self.subTest(msg=table.__name__):
                self.assertEqual(expected_name, table.get_table())
            # end with
        # end for
    # end def

    def test_table_name(self):
        tables = {
            SystemUnderTest: 'cool_table_yo',
            OtherTable: 'other_table',
            TheReferenceHasBeenDoubled: 'double_reference',
        }
        for table, expected_name in tables.items():
            with self.subTest(msg=table.__name__):
                self.assertEqual(expected_name, table.get_name())
            # end with
        # end for
    # end def

    def test_primary_keys(self):
        tables = {
            SystemUnderTest: ['t0_id'],
            OtherTable: ['id_part_1', 'id_part_2'],
            TheReferenceHasBeenDoubled: ['another_reference'],
        }
        for table, expected_name in tables.items():
            with self.subTest(msg=table.__name__):
                self.assertEqual(expected_name, table.get_primary_keys_keys())
            # end with
        # end for
    # end def

    def test_ignored_fields(self):
        tables = {
            SystemUnderTest: ['_table_name', '_ignored_fields', '_automatic_fields', '_primary_keys', '_database_cache', '__selectable_fields', f'_{SystemUnderTest.__name__!s}__selectable_fields', '__slots__'],
            OtherTable: ['_table_name', '_ignored_fields', '_automatic_fields', '_primary_keys', '_database_cache', '__selectable_fields', f'_{OtherTable.__name__!s}__selectable_fields', '__slots__'],
            TheReferenceHasBeenDoubled: ['_table_name', '_ignored_fields', '_automatic_fields', '_primary_keys', '_database_cache', '__selectable_fields', f'_{TheReferenceHasBeenDoubled.__name__!s}__selectable_fields', '__slots__'],
        }
        for table, expected_name in tables.items():
            with self.subTest(msg=table.__name__):
                self.assertEqual(expected_name, table.get_ignored_fields())
            # end with
        # end for
    # end def

    def test_automatic_fields(self):
        tables = {
            SystemUnderTest: ['_table_name', '_ignored_fields', '_automatic_fields', '_primary_keys', '_database_cache', '__selectable_fields', f'_{SystemUnderTest.__name__!s}__selectable_fields', '__slots__'],
            OtherTable: ['_table_name', '_ignored_fields', '_automatic_fields', '_primary_keys', '_database_cache', '__selectable_fields', f'_{OtherTable.__name__!s}__selectable_fields', '__slots__'],
            TheReferenceHasBeenDoubled: ['_table_name', '_ignored_fields', '_automatic_fields', '_primary_keys', '_database_cache', '__selectable_fields', f'_{TheReferenceHasBeenDoubled.__name__!s}__selectable_fields', '__slots__'],
        }
        for table, expected_name in tables.items():
            with self.subTest(msg=table.__name__):
                self.assertEqual(expected_name, table.get_ignored_fields())
            # end with
        # end for
    # end def
# end class

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fastorm']

package_data = \
{'': ['*']}

install_requires = \
['asyncpg', 'luckydonald-utils>=0.83', 'typeguard', 'pydantic']

setup(name='fastorm',
      version='0.0.8',
      description='FastORM framework, easy to learn, fast to code',
      author='luckydonald',
      author_email='fastorm+code@luckydonald.de',
      url='https://github.com/luckydonald/fastorm',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.10.0',
     )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import dataclasses
import typing

import pydantic.fields
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

from pydantic.dataclasses import dataclass

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if


def __getitem__(self, key):
    return getattr(self, key)
# end def


def __iter__(self):
    return iter(dataclasses.astuple(self))
# end def


FIELD_REFERENCE_TYPE = typing.TypeVar("FIELD_REFERENCE_TYPE")

FIELD_REFERENCE_ITEM_WITH_TYPE = typing.Union[type, typing.Type['FastORM']]   # the FIELD_REFERENCE_TYPE of TYPE type()
FIELD_REFERENCE_ITEM_WITH_TYPE_HINT = pydantic.fields.FieldInfo  # the FIELD_REFERENCE_TYPE of TYPE pydantic.typehint


@dataclass
class FieldItem(typing.Generic[FIELD_REFERENCE_TYPE]):
    field: str
    type_: FIELD_REFERENCE_TYPE

    __getitem__ = __getitem__  # reuse, as it's the same function basically
    __iter__ = __iter__  # reuse, as it's the same function basically
# end class


@dataclass
class FieldInfo(typing.Generic[FIELD_REFERENCE_TYPE]):
    is_primary_key: bool
    types: typing.List[FieldItem[FIELD_REFERENCE_TYPE]]

    @property
    def resulting_type(self) -> FIELD_REFERENCE_TYPE:
        """
        The last type in the type resolving list.

        That means this is the final type this field is actually having after resolving all the table references.
        TL;DR: flattened to the end type
        """
        return self.types[-1].type_
    # end def

    @property
    def referenced_type(self) -> FIELD_REFERENCE_TYPE:
        """
        The first type in the type resolving list.

        That means this is the first type, pointing to either the actual type if there's no reference or the table it references to
        TL;DR: no flattening to the end type
        """
        return self.types[0].type_
    # end def

    @property
    def referenced_field(self) -> FIELD_REFERENCE_TYPE:
        """
        The first type in the type resolving list.

        That means this is the first type, pointing to either the actual type if there's no reference or the table it references to
        TL;DR: no flattening to the end type
        """
        return self.types[1].field
    # end def

    __getitem__ = __getitem__  # reuse, as it's the same function basically
    __iter__ = __iter__  # reuse, as it's the same function basically
# end class
